"""
Fresh approach to having types in Pedal.

We will use inheritance to share behavior, not to check identity.
Instead we'll have our own set of fields to handle things like that.

Polymorphism, subtyping, there's a lot to unpack here.

Where possible, let's try to interface with `typing`, shall we?


Still to do:
 - [X] specify_type
 - [ ] convert_type
 - [ ] TypeUnion
 - [X] DictType
 - [X] FileType
 - [X] ModuleType
 - [X] GeneratorType
 - [X] SetType, FrozenSetType
 - [ ] LiteralTuple
 - [ ] Binary operations
 - [ ] Unary operations
 - [X] Orderable operations
 - [X] Indexing operations
 - [X] Iteration operations
 - [X] Membership operations

 - [ ] RecursedType

 - [*] Finish function definitions
 - [ ] Finish container literals (do condensing)
 - [ ] Finish Subscripting
 - [X] Unary Ops
 - [X] Finish collection loops
 - [X] Finish While
 - [X] Finish With
 - [ ] Finish evaluate_type
 - [ ] Finish built-in constructor types

"""
from pedal.utilities.text import join_list_with_and, add_indefinite_article


BUILTIN_MODULES = {}
BUILTIN_NAMES = {}

_MODULE_LOADERS = {}


def register_builtin_module(name, module_function):
    _MODULE_LOADERS[name] = module_function


# TODO: Should tie type system into Report
def reset_builtin_modules():
    BUILTIN_MODULES.clear()
    for name, module_function in _MODULE_LOADERS.items():
        BUILTIN_MODULES[name] = module_function()


def void_definition(tifa, function, callee, arguments, named_arguments, location):
    return NoneType()


def identity_definition(tifa, function, callee, arguments, named_arguments, location):
    if function and function.the_self:
        # If method, return self
        return function.the_self
    elif arguments:
        # Get the first argument
        return arguments[0]
    elif named_arguments:
        # Get the first named argument as the key
        first_key, first_value = next(iter(named_arguments))
        # If **kwargs, then retrieve first arbitrary key
        if first_key is None:
            _, first_kwarg_value = next(iter(first_value))
            return first_kwarg_value
        return first_value
    return ImpossibleType()


def element_definition(tifa, function, callee, arguments, named_arguments, location):
    """ Element function definition """
    return arguments[0].index(IntType()) if arguments else ImpossibleType()


class Type:
    # Class static
    name: str
    singular_name: str
    plural_name: str
    immutable: bool
    # Per-Class and per-instance
    fields: dict
    parents: list
    # Default Universal Fields
    is_empty = False
    orderable = frozenset()

    def __init__(self):
        self.fields = self.fields.copy()

    def clone(self):
        """ Make a new instance of this Type based on the old instance """
        return self.__class__()

    def shallow_clone(self):
        return self.clone()

    def __str__(self):
        return str(self.__class__.__name__)

    def precise_description(self):
        """
        Returns: A human-friendly but very precise description of this type
        """
        return self.singular_name

    def clone_mutably(self):
        """ Makes a new instance IF the type is immutable; otherwise returns this instance. """
        if self.immutable:
            return self.clone()
        else:
            return self

    def set_index(self, index, value):
        return ImpossibleType()

    def index(self, key):
        """ Access a specific element given the key """
        return ImpossibleType()

    def allows_membership(self, key):
        return False

    def iterate(self):
        """ Returns a single type representing the iteration variable """
        return ImpossibleType()

    def break_apart(self):
        """ Returns a generator of the types as requested """
        while True:
            yield ImpossibleType()

    def get_attr(self, field: str):
        value = self.find_field(field, set())
        if value is None:
            return ImpossibleType()
        return value

    def add_attr(self, field: str, value):
        self.fields[field] = value

    def find_field(self, field: str, seen: set):
        if field in self.fields:
            return self.fields[field]
        seen.add(self)
        for parent in self.parents:
            if parent not in seen:
                result = parent.find_field(field, seen)
                if result:
                    return result
        return

    def list_fields(self, seen=None):
        if seen is None:
            seen = set()
        for field in self.fields:
            yield field
        seen.add(self)
        for parent in self.parents:
            if parent not in seen:
                yield from parent.list_fields(seen)

    def is_subtype(self, other, seen, indent=0):
        if isinstance(self, AnyType) or isinstance(other, AnyType):
            return True
        if type(self) == type(other):
            return True
        if self in seen:
            return True
        seen.add(self)
        return any(parent.is_subtype(other, seen, 1+indent) for parent in self.parents)

    def specify_subtype(self, receiving_type):
        if isinstance(receiving_type, AnyType):
            return self
        return receiving_type

    def add_type_arguments(self, type_arguments):
        return self

    @staticmethod
    def definition(tifa, function, callee, arguments, named_arguments, location):
        return ImpossibleType()

    def as_type(self, tifa=None, location=None):
        return self

    def is_equal(self, other):
        """ DEPRECATED """
        import pedal.types.normalize as normalize
        return is_subtype(self, normalize.normalize_type(other).as_type())


class AnyType(Type):
    """
    A special type used to indicate an unknown type.

    This is our Top Type!
    """
    name = "Unknown"
    singular_name = "an unknown value"
    plural_name = "unknown values"
    immutable = False
    fields = {}
    parents = []

    def set_index(self, index, value):
        return AnyType()

    def index(self, key):
        """ Access a specific element given the key """
        return AnyType()

    def allows_membership(self, key):
        return True

    def iterate(self):
        """ Returns a single type representing the iteration variable """
        return AnyType()

    def break_apart(self):
        """ Returns a generator of the types as requested """
        while True:
            yield AnyType()

    def get_attr(self, field: str):
        return AnyType()


class ImpossibleType(Type):
    """
    A special type used to indicate a type situation that is impossible.
    The real type would be unknowable because of invalid code.

    This is our Bottom Type!
    """
    name = "Impossible"
    singular_name = "an impossible value"
    plural_name = "impossible value"
    immutable = False
    fields = {}
    parents = []


class NoneType(Type):
    name = "NoneType"
    singular_name = "a None"
    plural_name = "None"
    immutable = True
    fields = {}
    parents = []


class TypeUnion(Type):
    """
    TODO: Decide if using!
    A type that could be one of several possible types
    """
    name = "Type Union"
    singular_name = "one of several possible types"
    plural_name = "several possible types"
    immutable = False
    fields = {}
    parents = []

    def __init__(self, possible_types):
        super().__init__()
        self.possible_types = set(possible_types)


class FunctionType(Type):
    name = 'Function'
    singular_name = 'a function'
    plural_name = 'functions'
    immutable = True
    fields = {}
    parents = []
    #: the_self: Type | None

    def __init__(self, name="*Anonymous", definition=None, returns=None, the_self=None):
        super().__init__()
        if definition is None:
            if returns is None or returns == 'void':
                definition = void_definition
            elif returns == 'identity':
                definition = identity_definition
            elif returns == 'element':
                definition = element_definition
            else:
                def definition(tifa, function, callee, arguments, named_arguments, location):
                    """ Generic function """
                    return returns().clone()
        self.definition = definition
        self.name = name
        self.returns = returns
        self.the_self = the_self

    def clone(self):
        """ Create a deep copy of this function type. """
        returns = self.returns
        the_self = self.the_self if self.the_self is None else self.the_self.clone()
        return FunctionType(name=self.name, definition=self.definition, returns=returns, the_self=the_self)

    def clone_mutably(self):
        returns = self.returns
        the_self = self.the_self if self.the_self is None else self.the_self.clone_mutably()
        return FunctionType(name=self.name, definition=self.definition, returns=returns, the_self=the_self)

    def as_method(self, the_self):
        return FunctionType(name=self.name, definition=self.definition, returns=self.returns, the_self=the_self)


class NumType(Type):
    name = "Num"
    singular_name = "a number"
    plural_name = "numbers"
    immutable = True
    fields = {}
    parents = []


class IntType(Type):
    name = "Integer"
    singular_name = "an integer"
    plural_name = "integers"
    immutable = True
    fields = {}
    parents = [NumType()]

    def is_subtype(self, other, seen, indent=0):
        if isinstance(other, LiteralInt):
            return True
        return super().is_subtype(other, seen, indent)


class FloatType(Type):
    name = "Float"
    singular_name = "a float"
    plural_name = "floats"
    immutable = True
    fields = {}
    parents = [NumType()]

    def is_subtype(self, other, seen, indent=0):
        if isinstance(other, LiteralFloat):
            return True
        return super().is_subtype(other, seen, indent)


class BoolType(Type):
    name = "Boolean"
    singular_name = "a boolean"
    plural_name = "booleans"
    immutable = True
    fields = {}
    parents = []

    def is_subtype(self, other, seen, indent=0):
        if isinstance(other, LiteralBool):
            return True
        return super().is_subtype(other, seen, indent)


class ExceptionType(Type):
    name = "Exception"
    singular_name = "an exception"
    plural_name = "exceptions"
    immutable = False
    fields = {}
    parents = []


class StrType(Type):
    name = "String"
    singular_name = "a string"
    plural_name = "strings"
    immutable = True
    # Local
    is_empty: bool
    fields = {}
    parents = []

    def __init__(self, is_empty=False):
        super().__init__()
        self.is_empty = is_empty

    def index(self, key):
        # TODO: Support a flag for whether Numbers are allowed for Integers
        return StrType(False) if is_subtype(key, IntType()) else ImpossibleType()

    def is_subtype(self, other, seen, indent=0):
        if isinstance(other, LiteralStr):
            return True
        return super().is_subtype(other, seen, indent)

    def iterate(self):
        return StrType(False)

    def break_apart(self):
        while True:
            yield StrType(False)

    def allows_membership(self, key):
        return is_subtype(key, StrType())


class ElementContainerType(Type):
    is_empty: bool
    element_type: Type
    _singular_names: tuple
    _plural_names: tuple

    def __init__(self, is_empty=True, element_type=None):
        super().__init__()
        self.is_empty = is_empty
        if element_type is None:
            element_type = AnyType()
        self.element_type = element_type

    def __str__(self):
        # TODO: Handle other recursively defined definitions and nested recursion
        if self == self.element_type:
            return f"Recursively-defined {'Empty' if self.is_empty else ''}{self.name}Type"
        return f"{'Empty' if self.is_empty else ''}{self.name}Type[{self.element_type}]"

    @property
    def singular_name(self):
        descr = self._singular_names[self.is_empty]
        if not isinstance(self.element_type, AnyType):
            descr += " of " + self.element_type.plural_name
        return descr

    @property
    def plural_name(self):
        descr = self._plural_names[self.is_empty]
        if not isinstance(self.element_type, AnyType):
            descr += " of " + self.element_type.plural_name
        return descr

    def clone(self):
        return self.__class__(self.is_empty, self.element_type.clone())

    def shallow_clone(self):
        return self.__class__(self.is_empty, self.element_type)

    def is_subtype(self, other, seen, indent=0):
        if isinstance(self, AnyType) or isinstance(other, AnyType):
            return True
        if self in seen:
            return True
        seen.add(self)
        if type(self) == type(other) or any(parent.is_subtype(other, seen, indent+1) for parent in self.parents):
            other_element_type = other.element_type
            result = self.element_type.is_subtype(other_element_type, seen, indent+1)
            return result
        return False

    def specify_subtype(self, receiving_type):
        if isinstance(receiving_type, AnyType):
            return self
        if type(self) == type(receiving_type):
            receiving_type.element_type = self.element_type.specify_subtype(receiving_type.element_type)
            return receiving_type

    def add_type_arguments(self, type_arguments):
        self.element_type = (type_arguments[0] if len(type_arguments) == 1
                             else TupleType(type_arguments))
        return self

    def break_apart(self):
        while True:
            yield self.element_type

    def iterate(self):
        return self.element_type

    def allows_membership(self, key):
        return is_subtype(key, self.element_type)

    def as_type(self, tifa=None, location=None):
        return self.__class__(self.is_empty, self.element_type.as_type(tifa, location))


class SetType(ElementContainerType):
    name = "Set"
    immutable = False
    fields = {}
    parents = []
    _singular_names = ("a set", "an empty set")
    _plural_names = ("sets", "empty sets")

    @staticmethod
    def add(tifa, function, callee, arguments, named_arguments, location):
        if len(arguments) == 1:
            if not function or not function.the_self:
                raise ValueError("Attempted to call add method statically")
            self_type = function.the_self
            self_type.element_type = arguments[0]
            self_type.is_empty = False
            if callee:
                tifa.append_variable(callee.name, callee.type, location)
        return NoneType()


class FrozenSetType(ElementContainerType):
    name = "FrozenSet"
    immutable = True
    fields = {}
    parents = []
    _singular_names = ("a frozen set", "an empty, frozen set")
    _plural_names = ("frozen sets", "empty frozen sets")


class ListType(ElementContainerType):
    name = "List"
    immutable = False
    # Local
    fields = {}
    parents = []
    _singular_names = ("a list", "an empty list")
    _plural_names = ("lists", "empty lists")
    # TODO: Add in support for "protected_fields"
    protected_fields = ["append"]

    def index(self, key):
        # TODO: Support a flag for whether Numbers are allowed for Integers
        return self.element_type if is_subtype(key, IntType()) else ImpossibleType()

    @staticmethod
    def append(tifa, function, callee, arguments, named_arguments, location):
        if len(arguments) == 1:
            if not function or not function.the_self:
                raise ValueError("Attempted to call append method statically")
            self_type = function.the_self
            if not is_subtype(self_type.element_type, arguments[0]):
                tifa._issue('type_change_append', location,
                            self_type.element_type, arguments[0])
            self_type.element_type = arguments[0]
            self_type.is_empty = False
            if callee:
                tifa.append_variable(callee.name, callee.type, location)
        return NoneType()

    # TODO: extend function!


class GeneratorType(ElementContainerType):
    name = "Generator"
    immutable = True
    # Local
    fields = {}
    parents = []
    _singular_names = ("a generator", "an empty generator")
    _plural_names = ("generators", "empty generators")

    def index(self, key):
        return ImpossibleType()


class TupleType(Type):
    name = "Tuple"
    immutable = True
    # Local
    is_empty: bool
    element_types: tuple #: tuple[Type]
    fields = {}
    parents = []

    def __init__(self, element_types=None):
        super().__init__()
        if element_types is None:
            element_types = tuple()
        self.element_types = element_types
        self.is_empty = not element_types

    @property
    def singular_name(self):
        if self.is_empty:
            descr = "an empty tuple"
        else:
            descr = "a tuple"
        if not isinstance(self.element_types, AnyType):
            descr += " of " + join_list_with_and(t.singular_name for t in self.element_types)
        return descr

    @property
    def plural_name(self):
        if self.is_empty:
            descr = "empty tuples"
        else:
            descr = "tuples"
        if not isinstance(self.element_types, AnyType):
            descr += " of " + join_list_with_and(t.singular_name for t in self.element_types)
        return descr

    def clone(self):
        return TupleType(tuple(t.clone() for t in self.element_types))

    def shallow_clone(self):
        return TupleType(tuple(self.element_types))

    def index(self, key):
        # TODO: Support a flag for whether Numbers are allowed for Integers
        # TODO: Return a TypeUnion instead of the first one
        if isinstance(key, LiteralInt):
            if key.value >= len(self.element_types):
                return ImpossibleType()
            return self.element_types[key.value]
        elif is_subtype(key, IntType()) and self.element_types:
            return self.element_types[0]
        elif isinstance(key, int):
            if key >= len(self.element_types):
                return ImpossibleType()
            return self.element_types[key]
        else:
            return ImpossibleType()

    def iterate(self):
        if len(self.element_types) == 0:
            return AnyType()
        potential = widest_type(self.element_types)
        if potential is not None:
            return potential
        return TypeUnion(self.element_types)

    def break_apart(self):
        for t in self.element_types:
            yield t
        while True:
            yield ImpossibleType()

    def allows_membership(self, key):
        return any(is_subtype(key, t) for t in self.element_types)

    def is_subtype(self, other, seen, indent=0):
        if isinstance(self, AnyType) or isinstance(other, AnyType):
            return True
        if self in seen:
            return True
        seen.add(self)
        if type(self) == type(other) or any(parent.is_subtype(other, seen, indent + 1) for parent in self.parents):
            return all(e.is_subtype(e2, seen, indent+1)
                       for e, e2 in zip(self.element_types, other.element_types))
        return False

    def specify_subtype(self, receiving_type):
        if isinstance(receiving_type, AnyType):
            return self
        if type(self) == type(receiving_type) and len(self.element_types) == len(receiving_type.element_types):
            receiving_type.element_types = tuple((e.specify_subtype(e2)
                                                  for e, e2 in zip(self.element_types,
                                                                   receiving_type.element_types)))
            return receiving_type

    def add_type_arguments(self, type_arguments):
        self.element_types = type_arguments
        return self

    def as_type(self, tifa=None, location=None):
        return TupleType([t.as_type(tifa, location) for t in self.element_types])


class DictType(Type):
    name = "Dictionary"
    immutable = False
    # Local
    fields = {}
    parents = []
    _singular_names = ("a dictionary", "an empty dictionary")
    _plural_names = ("dictionaries", "empty dictionaries")
    element_types: list

    def __init__(self, element_types=None):
        super().__init__()
        if element_types is None:
            element_types = []
        self.element_types = element_types

    def _find_best_item(self, key):
        if isinstance(key, LiteralValue):
            for index, (potential_key, value) in enumerate(self.element_types):
                if isinstance(potential_key, LiteralValue):
                    if potential_key.value == key.value:
                        return index, potential_key, value
        candidates = [(index, potential_key, value)
                      for index, (potential_key, value) in enumerate(self.element_types)
                      if is_subtype(key, potential_key)]
        # TODO: Have some sense of "best" and "worst" subtype relationships?
        return candidates[0] if candidates else (None, None, None)


    @property
    def is_empty(self):
        return not self.element_types

    @property
    def singular_name(self):
        base = self._singular_names[self.is_empty]
        if len(self.element_types) == 1:
            key, value = next(iter(self.element_types))
            base += f" mapping {key.plural_name} to {value.plural_name}"
        elif len(self.element_types) > 1:
            base += " mapping " + join_list_with_and(f"{key.singular_name} to {value.plural_name}"
                                                     for key, value in self.element_types)
        return base

    @property
    def plural_name(self):
        base = self._plural_names[self.is_empty]
        if len(self.element_types) == 1:
            key, value = next(iter(self.element_types))
            base += f" mapping {key.plural_name} to {value.plural_name}"
        elif len(self.element_types) > 1:
            base += " mapping " + join_list_with_and(f"{key.singular_name} to {value.plural_name}"
                                                     for key, value in self.element_types)
        return base

    def clone(self):
        return DictType([(key.clone(), value.clone()) for key, value in self.element_types])

    def shallow_clone(self):
        return DictType(list(self.element_types))

    def is_subtype(self, other, seen, indent=0):
        if isinstance(self, AnyType) or isinstance(other, AnyType):
            return True
        if self in seen:
            return True
        seen.add(self)
        # Allows extra keys in the other, but must have at least all the expected keys/values
        if type(self) == type(other) or any(parent.is_subtype(other, seen, indent + 1) for parent in self.parents):
            return all(
                any(
                    self_key.is_subtype(other_key, seen.copy(), indent + 1)
                    and self_value.is_subtype(other_value, seen.copy(), indent + 1)
                    for other_key, other_value in other.element_types
                ) for self_key, self_value in self.element_types)
        return False

    def add_type_arguments(self, type_arguments):
        self.element_types = [type_arguments]
        return self

    def index(self, key):
        # TODO: Support a flag for whether Numbers are allowed for Integers
        if self.is_empty:
            return ImpossibleType()
        return self._find_best_item(key)[2] or ImpossibleType()

    def allows_membership(self, key):
        return any(is_subtype(key, potential_key) for potential_key, value in self.element_types)

    def iterate(self):
        keys = [k for k, v in self.element_types]
        if len(keys) == 0:
            return AnyType()
        elif len(keys) == 1:
            return keys[0]
        else:
            potential = widest_type(keys)
            if potential:
                return potential
            return TypeUnion(keys)

    def break_apart(self):
        for key, value in self.element_types:
            yield key
        while True:
            yield ImpossibleType()

    def set_index(self, key, value_type):
        candidate = self._find_best_item(key)
        if candidate[0] is None:
            self.element_types.append((key, value_type))
        else:
            if isinstance(key, LiteralValue):
                if isinstance(candidate[1], LiteralValue):
                    if key.value == candidate[1].value:
                        self.element_types[candidate[0]] = candidate[1:]
                self.element_types.append((key, value_type))
            else:
                self.element_types[candidate[0]] = (widen_type(key, candidate[1]),
                                                    widen_type(value_type, candidate[2]))

    def as_type(self, tifa=None, location=None):
        return DictType([(k.as_type(tifa, location), v.as_type(tifa, location))
                         for k, v in self.element_types])

    @staticmethod
    def keys(tifa, function, callee, arguments, named_arguments, location):
        self_type = function.the_self
        if self_type.is_empty:
            return GeneratorType(True)
        return GeneratorType(False, widest_type([key for key, value in self_type.element_types]))

    @staticmethod
    def values(tifa, function, callee, arguments, named_arguments, location):
        self_type = function.the_self
        if self_type.is_empty:
            return GeneratorType(True)
        return GeneratorType(False, widest_type([value for key, value in self_type.element_types]))

    @staticmethod
    def items(tifa, function, callee, arguments, named_arguments, location):
        self_type = function.the_self
        if self_type.is_empty:
            return GeneratorType(True)
        keys, values = [], []
        for key, value in self_type.element_types:
            keys.append(key)
            values.append(value)
        return GeneratorType(False, TupleType([widest_type(keys), widest_type(values)]))

    @staticmethod
    def get(tifa, function, callee, arguments, named_arguments, location):
        self_type = function.the_self
        if not arguments:
            return ImpossibleType()
        return self_type._find_best_item(arguments[0])[1] or NoneType()

    @staticmethod
    def copy(tifa, function, callee, arguments, named_arguments, location):
        return function.the_self.clone()

    pop = get

    # TODO: popitem, setdefault, update


class ClassType(Type):
    name = "Class"
    singular_name = "a class"
    plural_name = "classes"
    immutable = False
    fields = {}
    parents = []

    def __init__(self, name, fields=None, parents=None):
        super().__init__()
        self.name = name
        self.fields = fields if fields is not None else {}
        self.parents = parents if parents is not None else []

    @property
    def singular_name(self):
        return f"a {self.name} class"

    @property
    def plural_name(self):
        return f"{self.name} classes"

    def __str__(self):
        return f"ClassType[{self.name}]"

    def get_constructor(self):
        """ We actually follow the true constructor/initializer separation here!"""
        # Default constructor
        def constructor(*args):
            return InstanceType(self)

        return FunctionType(name='__init__', definition=constructor)

    def clone(self):
        return ClassType(self.name, self.fields.copy(), self.parents().copy())

    def is_subtype(self, other, seen, indent=0):
        if isinstance(self, AnyType) or isinstance(other, AnyType):
            return True
        if self in seen:
            return True
        seen.add(self)
        if other == TYPE_TYPE or self == other:
            return True
        return any(parent.is_subtype(other, seen, 1+indent) for parent in self.parents)

    def as_type(self, tifa=None, location=None):
        return InstanceType(self)


class BuiltinConstructorType(Type):
    name = "type"
    immutable = True
    fields = {}
    parents = []
    type_arguments: tuple  # type: tuple[Type]

    def __init__(self, name=None, type_arguments=None):
        super().__init__()
        if name is not None:
            self.name = name
        self.type_arguments = type_arguments

    def clone(self):
        return self.__class__(self.name, self.type_arguments)

    @property
    def singular_name(self):
        return f'the {self.name} constructor function'

    @property
    def plural_name(self):
        return f'{self.name} constructor functions'

    def index(self, key):
        self.type_arguments = key
        return self

    def definition(self, tifa, function, callee, arguments, named_arguments, location):
        return AnyType()

    def as_type(self, tifa=None, location=None):
        return self.definition(tifa, self, None, [], [], location)

    def add_type_arguments(self, type_arguments):
        self.type_arguments = type_arguments[0] if len(type_arguments) == 1 else TupleType(type_arguments)
        return self

class IntConstructor(BuiltinConstructorType):
    name = 'int'
    definition = lambda *args: IntType()


class FloatConstructor(BuiltinConstructorType):
    name = 'float'
    definition = lambda *args: FloatType()


class BoolConstructor(BuiltinConstructorType):
    name = 'bool'
    definition = lambda *args: BoolType()


class StrConstructor(BuiltinConstructorType):
    name = 'str'

    def definition(self, tifa, function, callee, arguments, named_arguments, location):
        if arguments:
            if is_subtype(arguments[0], StrType()):
                return arguments[0].clone()
            return StrType(False)
        return StrType(True)


class ElementContainerConstructor(BuiltinConstructorType):
    element_container_type: type(ElementContainerType)

    def definition(self, tifa, function, callee, arguments, named_arguments, location):
        element_type = self.type_arguments
        #if isinstance(element_type, BuiltinConstructorType):
        #    element_type = element_type.definition(self, element_type, None, [], [], location)
        if element_type is None:
            element_type = AnyType()
        element_type = element_type.as_type(tifa, location)
        result_type = self.element_container_type(bool(arguments), element_type)
        if arguments:
            result_type = specify_subtype(result_type, arguments[0].iterate())
        return result_type


class ListConstructor(ElementContainerConstructor):
    name = 'list'
    element_container_type = ListType


class SetConstructor(ElementContainerConstructor):
    name = 'set'
    element_container_type = SetType


class FrozenSetConstructor(ElementContainerConstructor):
    name = 'frozenset'
    element_container_type = FrozenSetType


class DictConstructor(BuiltinConstructorType):
    name = 'dict'

    def definition(self, tifa, function, callee, arguments, named_arguments, location):
        element_type = self.type_arguments
        if element_type is None:
            element_type = AnyType()
        key_type, value_type = element_type.index(LiteralInt(0)), element_type.index(LiteralInt(1))
        if isinstance(key_type, BuiltinConstructorType):
            key_type = key_type.definition(self, key_type, None, [], [], location)
        if isinstance(value_type, BuiltinConstructorType):
            value_type = value_type.definition(self, value_type, None, [], [], location)
        if arguments:
            # TODO: Parse the argument for more data!
            return DictType([(key_type, value_type)])
        return DictType([(key_type, value_type)])


class TupleConstructor(BuiltinConstructorType):
    name = 'tuple'

    def definition(self, tifa, function, callee, arguments, named_arguments, location):
        element_type = self.type_arguments
        if element_type is None:
            # TODO: How do we handle completely generic tuples?
            # For now, we'll treat them as if they were empty tuples...
            return TupleType([])
        # Assume single elements are actually just a tuple of length one
        if not isinstance(element_type, (tuple, set, list)):
            element_type = [element_type]
        element_type = [e.definition(self, e, None, [], [], location)
                        for e in element_type if isinstance(e, BuiltinConstructorType)]
        result_type = TupleType(element_type)
        if arguments:
            result_type = specify_subtype(result_type, arguments[0])
        return result_type


TYPE_TYPE = BuiltinConstructorType("type")


class InstanceType(Type):
    immutable = False
    fields = {}

    def __init__(self, parent):
        super().__init__()
        self.parents = [parent]
        self.fields = {}
        self.name = parent.name

    def __str__(self):
        return "InstanceTypeOf" + str(self.parents[0].name)

    @property
    def singular_name(self):
        return add_indefinite_article(self.name)

    @property
    def plural_name(self):
        return self.name

    def clone(self):
        return InstanceType(self.parents[0])

    def is_subtype(self, other, seen, indent=0):
        if isinstance(self, AnyType) or isinstance(other, AnyType):
            return True
        # Avoid recursion
        if self in seen:
            return True
        seen.add(self)
        # Check that the structure matches the expected
        if type(self) == type(other) and other.name == self.name:
            for expected_field_name in set(other.list_fields()):
                if expected_field_name.startswith('__'):
                    continue
                expected_field_type = other.find_field(expected_field_name, set())
                actual_field_type = self.find_field(expected_field_name, set())
                if actual_field_type is None or expected_field_type is None:
                    return False
                if not is_subtype(actual_field_type, expected_field_type):
                    return False
            return True

        def check_parent(current):
            if isinstance(current, AnyType):
                return True
            if current.name == other.name:
                return True
            if current in seen:
                return True
            seen.add(self)
            return any(check_parent(parent) for parent in current.parents)
        result = any(check_parent(parent) for parent in self.parents)
        return result

    def add_attr(self, field: str, value):
        if field in self.fields:
            if not is_subtype(self.fields[field], value):
                return ImpossibleType()
        self.fields[field] = value
        return value


def make_dataclass(tifa, function, callee, arguments, named_arguments: dict, location):
    class_type = arguments[0]
    simple_class_fields = [(field, expected_type) for field, expected_type in class_type.fields.items()
                           if not isinstance(expected_type, FunctionType)]

    def constructor(tifa, function, callee, arguments, named_arguments: dict, location):
        seen_fields = set()
        missing_fields = []
        arguments = arguments.copy()
        the_self = arguments.pop(0) if arguments else None
        named_arguments = named_arguments.copy()
        new_instance = InstanceType(class_type)
        # TODO: Handle kwargs better
        kwargs = [(k, v) for k, v in named_arguments if k is None]
        for field, expected_type in simple_class_fields:
            # Key is found?
            # Key's value has right type?
            if field in named_arguments:
                value = named_arguments.pop(field)
            elif field in kwargs:
                value = kwargs.pop(field)
            elif arguments:
                value = arguments.pop(0)
            else:
                #tifa._issue('missing_parameter', location, field, expected_type)
                missing_fields.append((field, expected_type))
                continue
            seen_fields.add(field)
            if not is_subtype(value, expected_type):
                tifa._issue('field_type_mismatch', location, field, expected_type, value)
            new_instance.add_attr(field, value.clone_mutably())
        expected_field_count = len(simple_class_fields)
        actual_field_count = len(seen_fields) + len(arguments) + len(named_arguments) + len(kwargs)
        if expected_field_count != actual_field_count:
            # TODO: unexpected_dataclass_parameters
            tifa._issue('incorrect_arity', location, class_type.name, expected_field_count, actual_field_count, True)
        return new_instance
    constructor_function_type = FunctionType('__init__', definition=constructor)
    class_type.add_attr('__init__', constructor_function_type)
    return class_type


class ModuleType(Type):
    name = "Module"
    singular_name = "a module"
    plural_name = "modules"
    immutable = False
    fields = {}
    parents = []
    submodules: dict

    def __init__(self, name, fields, submodules=None):
        super().__init__()
        self.name = name
        self.fields = fields
        self.submodules = submodules if submodules is not None else {}

    def clone(self):
        """ Create a copy of this ModuleType """
        submodules = {n: v.clone() for n, v in self.submodules.items()}
        fields = {n: v.clone() for n, v in self.fields.items()}
        return ModuleType(name=self.name, submodules=submodules, fields=fields)

    def shallow_clone(self):
        return ModuleType(name=self.name, submodules=self.submodules, fields=self.fields)

    def as_type(self, tifa=None, location=None):
        return ModuleType(name=self.name, submodules=self.submodules, fields={
            f: t.as_type(tifa, location) for f, t in self.fields.items()
        })


class FileType(Type):
    name = "FileType"
    singular_name = "a file"
    plural_name = "files"
    immutable = False
    fields = {}
    parents = []

    def iterate(self):
        return StrType()

    def break_apart(self):
        while True:
            yield StrType()


# TODO: Implement args/kwargs signature checking

NumType.fields = {
    # TODO: Finish these
    "as_integer_ratio": FunctionType("as_integer_ratio", returns=lambda: TupleType((IntType(), IntType())))
}

BoolType.fields = IntType.fields = {
    # TODO: Finish these
    "bit_length": FunctionType("bit_length", definition=IntType)
}

FloatType.fields = {
    "is_integer": FunctionType("is_integer", returns=BoolType)
}

StrType.fields = {
    # Methods that return strings
    "capitalize": FunctionType('capitalize', returns=StrType),
    "center": FunctionType(name='center', returns=StrType),
    "expandtabs": FunctionType(name='expandtabs', returns=StrType),
    "join": FunctionType(name='join', returns=StrType),
    "ljust": FunctionType(name='ljust', returns=StrType),
    "lower": FunctionType(name='lower', returns=StrType),
    "lstrip": FunctionType(name='lstrip', returns=StrType),
    "replace": FunctionType(name='replace', returns=StrType),
    "rjust": FunctionType(name='rjust', returns=StrType),
    "rstrip": FunctionType(name='rstrip', returns=StrType),
    "strip": FunctionType(name='strip', returns=StrType),
    "swapcase": FunctionType(name='swapcase', returns=StrType),
    "title": FunctionType(name='title', returns=StrType),
    "translate": FunctionType(name='translate', returns=StrType),
    "upper": FunctionType(name='upper', returns=StrType),
    "zfill": FunctionType(name='zfill', returns=StrType),
    # Methods that return numbers
    "count": FunctionType(name='count', returns=NumType),
    "find": FunctionType(name='find', returns=NumType),
    "rfind": FunctionType(name='rfind', returns=NumType),
    "index": FunctionType(name='index', returns=NumType),
    "rindex": FunctionType(name='rindex', returns=NumType),
    # Methods that return booleans
    "startswith": FunctionType(name='startswith', returns=BoolType),
    "endswith": FunctionType(name='endswith', returns=BoolType),
    "isalnum": FunctionType(name='isalnum', returns=BoolType),
    "isalpha": FunctionType(name='isalpha', returns=BoolType),
    "isdigit": FunctionType(name='isdigit', returns=BoolType),
    "islower": FunctionType(name='islower', returns=BoolType),
    "isspace": FunctionType(name='isspace', returns=BoolType),
    "istitle": FunctionType(name='istitle', returns=BoolType),
    "isupper": FunctionType(name='isupper', returns=BoolType),
    # Methods that return List of Strings
    "rsplit": FunctionType(name='rsplit', returns=lambda: ListType(False, StrType())),
    "split": FunctionType(name='split', returns=lambda: ListType(False, StrType())),
    "splitlines": FunctionType(name='splitlines', returns=lambda: ListType(False, StrType()))
}

SetType.fields = {
    "add": FunctionType("add", definition=SetType.add)
}

ListType.fields = {
    "append": FunctionType("append", definition=ListType.append),
}

TupleType.fields = {
    "count": FunctionType("count", returns=IntType),
    "index": FunctionType("index", returns=IntType)
}

DictType.fields = {
    'items': FunctionType('items', DictType.items),
    'keys': FunctionType('keys', DictType.keys),
    'values': FunctionType('values', DictType.values),
    'get': FunctionType('get', DictType.get),
    'copy': FunctionType('copy', DictType.copy),
    'pop': FunctionType('pop', DictType.pop),
    # TODO: Finish these
    # 'popitem': FunctionType('popitem', DictType.popitem),
    # 'setdefault': FunctionType('setdefault', DictType.setdefault),
    # 'update': FunctionType('update', DictType.update),
}

FileType.fields = {
    'close': FunctionType('close', definition=void_definition),
    'read': FunctionType('read', returns=StrType),
    'readline': FunctionType('readline', returns=StrType),
    'readlines': FunctionType('readlines', returns=lambda: ListType(False, StrType())),
    'write': FunctionType('write', returns=IntType),
    'writelines': FunctionType('writelines', returns=IntType),
    'tell': FunctionType('tell', returns=IntType),
    'seek': FunctionType('seek', returns=IntType),
}


class LiteralValue:
    """
    Special class flag that can be used to check if the type was defined as a literal.
    """
    value: any
    parents: list

    def __init__(self, value):
        self.value = value

    def promote(self):
        return self.parents[0]

    def clone(self):
        return self.__class__(self.value)


class LiteralBool(LiteralValue, BoolType):
    value: str
    singular_name = "a boolean"
    plural_name = "booleans"
    fields = {}
    parents = [BoolType()]


class LiteralStr(LiteralValue, StrType):
    value: str
    singular_name = "a string"
    plural_name = "strings"
    fields = {}

    @property
    def parents(self):
        return [StrType(self.is_empty)]

    def promote(self):
        return StrType(self.is_empty)


class LiteralInt(LiteralValue, Type):
    name = "LiteralInteger"
    singular_name = "an integer"
    plural_name = "integers"
    immutable = True
    value: int
    fields = {}
    parents = [IntType()]

    def __str__(self):
        return f"LiteralInteger[{self.value}]"


class LiteralFloat(LiteralValue, Type):
    name = "LiteralFloat"
    singular_name = "a float"
    plural_name = "floats"
    immutable = True
    value: int
    fields = {}
    parents = [FloatType()]

    def __str__(self):
        return f"LiteralFloat[{self.value}]"


# Types that support ordering
NUMERIC_TYPES = frozenset([
    NumType, IntType, FloatType, BoolType, LiteralInt, LiteralFloat
])
NumType.orderable = IntType.orderable = FloatType.orderable = BoolType.orderable = NUMERIC_TYPES
LiteralInt.orderable = LiteralFloat.orderable = NUMERIC_TYPES
StrType.orderable = LiteralStr.orderable = frozenset([StrType, LiteralStr])

for self_orderable in [ListType, SetType, TupleType]:
    self_orderable.orderable = frozenset([self_orderable])

TYPE_STRINGS = {
    "str": StrType, "string": StrType,
    "num": NumType, "number": NumType,
    "int": IntType, "integer": NumType, "float": FloatType,
    "complex": NumType,
    "bool": BoolType, "boolean": BoolType,
    "none": NoneType, "nonetype": NoneType,
    "dict": DictType, "dictionary": DictType,
    "list": ListConstructor,
    "tuple": TupleType,
    "set": SetConstructor,
    "file": FileType,
    "func": FunctionType, "function": FunctionType,
    "class": ClassType,
}


""" TOBEDEPRECATED: legacy support for using NumType and other similar names """
PEDAL_TYPE_NAMES = {
    "NumType": NumType,
    "StrType": StrType,
    "BoolType": BoolType,
    "ListType": ListType,
    "NoneType": NoneType
}


def exception_function(name):
    return FunctionType(name, returns=ExceptionType)


def int_function(name):
    return FunctionType(name, returns=IntType)


def float_function(name):
    return FunctionType(name, returns=FloatType)


def num_function(name):
    return FunctionType(name, returns=FloatType)


def bool_function(name):
    return FunctionType(name, returns=BoolType)


def void_function(name):
    return FunctionType(name, void_definition)


def str_function(name):
    return FunctionType(name, returns=StrType)


def is_subtype(left, right):
    if left is None or right is None:
        return False
    else:
        return left.is_subtype(right, seen=set())


def specify_subtype(specific_type, receiving_type):
    if specific_type is None or receiving_type is None:
        return None
    return specific_type.specify_subtype(receiving_type)


def widen_type(left, right):
    if is_subtype(right, left) and not is_subtype(left, right):
        return left
    elif is_subtype(left, right):
        return right
    else:
        return None


def widest_type(type_values):
    if not type_values:
        return
    first_type = type_values[0]
    for type_value in type_values[1:]:
        first_type = widen_type(first_type, type_value)
        if first_type is None:
            return
    return first_type


if __name__ == "__main__":
    # Test cases
    "Complex Subtypes"
    list_of_list_of_integers = ListType(False, ListType(False, IntType()))
    list_of_list_of_numbers = ListType(False, ListType(False, NumType()))
    list_of_integers = ListType(False, IntType())
    assert not is_subtype(LiteralInt(5), list_of_integers)
    assert is_subtype(ListType(True), list_of_list_of_integers)
    assert is_subtype(ListType(False, ListType(True)), list_of_list_of_integers)
    assert is_subtype(ListType(False, ListType(False, IntType())), list_of_list_of_integers)
    assert not is_subtype(ListType(False, ListType(False, NumType())), list_of_list_of_integers)
    assert is_subtype(ListType(False, ListType(False, IntType())), list_of_list_of_numbers)
    assert is_subtype(ListType(False, ListType(False, LiteralInt(5))), list_of_list_of_integers)
    assert not is_subtype(IntType(), list_of_list_of_integers)
    assert not is_subtype(IntType(), list_of_list_of_numbers)
    assert not is_subtype(NumType(), list_of_list_of_integers)
    assert not is_subtype(NumType(), list_of_list_of_numbers)
    assert not is_subtype(ListType(False, IntType()), list_of_list_of_integers)
    assert not is_subtype(ListType(False, NumType()), list_of_list_of_integers)
    print(list_of_list_of_integers, list_of_list_of_integers.singular_name)
    xyz_coordinates = TupleType((IntType, IntType, IntType))
    print(xyz_coordinates, xyz_coordinates.singular_name)
    list_of_xyz_coordinates = ListType(False, xyz_coordinates)
    print(list_of_xyz_coordinates, list_of_xyz_coordinates.singular_name)

    "Instances and Classes"
    Animal = ClassType("Animal")
    Dog = ClassType("Dog", parents=[Animal])
    Dog.add_attr("age", IntType())
    Dog.add_attr("fuzzy", BoolType())
    Dog.add_attr("name", StrType())
    dog_instances = Dog.get_constructor().definition()
    print(dog_instances)
    for field in ['name', 'age', 'fuzzy', 'is_pretty']:
        print("  ", field+":", dog_instances.get_attr(field))

    assert is_subtype(Dog, Dog), "Dogs are Dogs"

    "List of Instances"
    list_of_dog_instances = ListType(False, dog_instances)
    list_of_dogs = ListType(False, Dog)
    list_of_cats = ListType(False, ClassType("Cat"))
    list_of_animals = ListType(False, Animal)
    list_of_integers = ListType(False, IntType())
    assert is_subtype(list_of_dogs, ListType(False, TYPE_TYPE)), "Dog types are instances of Type Type"
    assert is_subtype(list_of_dog_instances, list_of_dogs), "Instances are subtypes of dogs"
    assert is_subtype(list_of_dog_instances, list_of_dog_instances), "Instances are subtypes of instances"
    assert is_subtype(list_of_dogs, list_of_dogs), "Dogs are not subtypes of dogs"
    assert not is_subtype(list_of_dog_instances, list_of_cats), "Instances of dogs are not subtypes of cats"
    assert is_subtype(list_of_dog_instances, list_of_animals), "Instances of dogs are subtypes of animals"
    assert not is_subtype(list_of_integers, list_of_animals), "Integers are not subtypes of animals"
    assert not is_subtype(StrType(), ClassType("Dog")), "Strings are not dogs"

    # Type widening
    assert type(widen_type(IntType(), LiteralInt(5))) == IntType
    assert type(widen_type(LiteralInt(3), IntType())) == IntType
    assert type(widen_type(NumType(), LiteralInt(5))) == NumType
    assert type(widen_type(LiteralInt(3), NumType())) == NumType
    assert type(widen_type(NumType(), IntType())) == NumType
    assert type(widen_type(IntType(), NumType())) == NumType
    assert widen_type(LiteralInt(3), StrType()) is None
    assert widen_type(StrType(), LiteralInt(3)) is None
    assert widen_type(dog_instances, Dog) == Dog
    assert widen_type(Dog, dog_instances) == Dog



    print("All good!")




import traceback
import unittest
import os
import sys
from textwrap import dedent
from pprint import pprint
from dataclasses import dataclass

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import pedal.tifa
import pedal.types.new_types as defs
import pedal.types.normalize as normalize
from pedal import contextualize_report, Submission, evaluate
from pedal.tifa import tifa_provide_module_type

unit_tests = {
    # Source Code, Shouldn't catch this, Should catch this
    'builtin_True': ['print(True)', ['initialization_problem'], []],
    'unread_variable': ['a = 0', [], ['unused_variable']],
    'undefined_variable': ['print(a)', [], ['initialization_problem']],
    'defined_read_variable': ['a = 0\nprint(a)', ['initialization_problem'], []],
    'overwritten_variable': ['a = 0\na = 5', [], ['overwritten_variable']],
    'unread_variables': ['a = 0\nb = 5', ['overwritten_variable'], ['unused_variable']],
    'wrwr_variable': ['a = [1]\nprint(a)\na = [1]\nprint(a)', [], []],
    # unconnected_blocks
    'unconnected_assign': ['a = ___', [], ['unconnected_blocks']],
    'unconnected_print': ['print(___)', [], ['unconnected_blocks']],

    'literal_in_call': ['print("dog" in input("test"))', [], []],
    'wrong_method_for_type': ['[].replace(",","")', [], []],

    "check_true_argument_in_function_call":
        ['def is_true(x:bool)->bool:\n return x\nis_true(True)', 'parameter_type_mismatch', ''],

    # Double call
    'double_call': ['def x(a):\n    return a\nx(5)\nx(3)', ['read_out_of_scope'], []],

    # Chained functions
    'chained_functions': ['def x():\n    return 0\ndef y():\n    x()\ny()',
                          ['read_out_of_scope', 'initialization_problem'], []],

    # String indexing and slicing
    'string_indexing_slicing': ['("a"[0] + ""[:])[:][0]', ['incompatible_types'], []],
    # List indexing and slicing
    'list_indexing_slicing': ['([0][0] + [1,2,3][:][2])', ['incompatible_types'], []],

    'returned_string':
        [
            'def pluralize(a_word):\n    return a_word+"s"\nnoun = pluralize("Dog")\nprint(noun + " can pet other " + noun)',
            ['incompatible_types'], []],
    'update_without_read':
        ['a = 0\na+= 1', ['initialization_problem'], ['unused_variable']],
    'update_and_read':
        ['a = 0\na+= 1\nprint(a)', ['initialization_problem', 'unused_variable'], []],
    'iterate_through_non_existing_list':
        ['for x in y:\n\tpass', ['unused_variable'], ['initialization_problem']],
    'iterate_through_list':
        ['y = [1,2,3]\nfor x in y:\n\tpass', ['unused_variable', 'initialization_problem'], []],
    'iterate_through_empty_list':
        ['y = []\nfor x in y:\n\tpass', ['unused_variable', 'initialization_problem'], ['iterating_over_empty_list']],
    'double_iterate_through_strings':
        ['ss = ["Testing", "Here"]\nfor a in ss:\n    print(a)\nfor b in a:\n    print(b)',
         ['iterating_over_non_list', 'iterating_over_empty_list'], []],
    'iterate_through_number':
        ['y = 5\nfor x in y:\n\tpass', ['unused_variable', 'initialization_problem'], ['iterating_over_non_list']],
    'iterate_over_iteration_variable':
        ['y = [1,2,3]\nfor y in y:\n\tpass', [], ['iteration_problem']],
    'type_change':
        ['a = 0\nprint(a)\na="T"\nprint(a)', [], ['type_changes']],
    'defined_in_if_root_but_not_other':
        ['if True:\n\ta = 0\nprint(a)', [], ['possible_initialization_problem']],
    'defined_in_both_branches':
        ['if True:\n\ta = 0\nelse:\n\ta = 1\nprint(a)', ['possible_initialization_problem'], []],
    'defined_in_else_root_but_not_other':
        ['if True:\n\tpass\nelse:\n\ta = 1\nprint(a)', [], ['possible_initialization_problem']],
    'defined_in_if_branch_but_others':
        ['if True:\n\tif False:\n\t\ta = 0\nprint(a)', [], ['possible_initialization_problem']],
    'defined_before_if_branch_but_not_others':
        ['if True:\n\ta = 0\nif False:\t\tpass\nprint(a)', [], ['possible_initialization_problem']],
    'defined_after_if_branch_but_not_others':
        ['if True:\n\tif False:\n\t\tpass\n\ta = 0\nprint(a)', [], ['possible_initialization_problem']],
    'defined_within_both_if_branches_but_not_others':
        ['if True:\n\tif False:\n\t\ta=0\n\telse:\n\t\ta = 0\nprint(a)', [], ['possible_initialization_problem']],
    'defined_in_all_branches':
        ['if True:\n\tif False:\n\t\ta=0\n\telse:\n\t\ta = 0\nelse:\n\ta=3\nprint(a)',
         ['possible_initialization_problem'], []],
    'read_in_if_branch_but_unset':
        ['if True:\n\tprint(a)', [], ['initialization_problem']],
    'read_in_else_branch_but_unset':
        ['if True:\n\tpass\nelse:\n\tprint(a)', [], ['initialization_problem']],
    'read_in_both_branches_but_unset':
        ['if True:\n\tprint(a)\nelse:\n\tprint(a)', [], ['initialization_problem']],
    'overwritten_in_both_branches':
        ['a = 0\nif True:\n\ta = 0\nelse:\n\ta = 1', [], ['overwritten_variable']],
    'overwritten_in_one_branch':
        ['a = 0\nif True:\n\tpass\nelse:\n\ta = 1', ['overwritten_variable'], []],
    'overwritten_in_inner_branch':
        ['a = 0\nif True:\n\tif False:\n\t\ta = 0\nelse:\n\ta = 1', ['overwritten_variable'], []],
    'overwritten_in_all_branch':
        ['a = 0\nif True:\n\tif False:\n\t\ta = 0\n\telse:\n\t\ta = 2\nelse:\n\ta = 1', [], ['overwritten_variable']],
    'overwritten_in_all_branches2':
        ['a = 0\nif True:\n\tprint(a)\n\tif False:\n\t\ta = 0\n\telse:\n\t\ta = 2\nelse:\n\ta = 1',
         ['overwritten_variable'], []],

    # TODO: This test case still fails! Investigate why, it's something quite complex.
    'take_pattern':
        ['def take(l):\n'
         ' t=True\n'
         ' r=[]\n'
         ' for i in l:\n'
         '  if i==".":\n'
         '    t=False\n'
         '  elif t:\n'
         '    r.append(i)\n'
         ' return r\n'
         'take([1,2,3, ".", 5, 6])', ['unused_variable', 'possible_initialization_problem', 'initialization_problem'], []],

    # TODO: This test case still fails! Investigate why, it's something quite complex.
    'possibly_overwritten_in_elif_branch':
        ['highest = 0\nfor score in [1,2]:\n    if False:\n        pass\n    elif False:\n        pass\n    else:\n        highest = 0\nhighest',
         ['possible_initialization_problem'], []],

    'andy_overwritten_in_elif_branch':
    ['from rates import bill_small,bill_medium,bill_large\n# do not delete the above line\nfrom cisc106 import assert_equal\n\ndef total_bill(data,base):\n    ''\n    if data < 100:\n        bill= bill_small(data,base)\n        return bill\n    elif data <= 1000 and data>100:\n        return bill_medium(data,base)\n    elif data > 1000:\n        bill = bill_large(data,base)\n        return float(bill)\n\nassert_equal(total_bill(30,80.),80.)\nassert_equal(total_bill(150,80.),95.5)\nassert_equal(total_bill(3000,80.0),160.)',
     ['possible_initialization_problem'], []],

    # Iterating over the result of a builtin
    'print_range':
        ['x = range(100)\nprint(x)', ['iterating_over_non_list'], []],
    'iterate_range':
        ['x = range(100)\nfor y in x:\n    print(y)', ['iterating_over_non_list'], []],
    'iterate_over_ranges_atomic_subtype':
        ['x = range(100)\nfor y in x:\n    pass\nfor z in y:\n    print(z)', [], ['iterating_over_non_list']],
    'iterate_over_split':
        ['for x in "a,b,c".split(","):\n  x+""', ['iterating_over_non_list', 'iterating_over_empty_list'], []],
    'iterate_over_string_upper':
        ['for l in "abc".upper():\n  l+""', ['iterating_over_non_list', 'iterating_over_empty_list'], []],

    # incompatible_types
    'add_int_str':
        ['a = 5 + "ERROR"', [], ['incompatible_types']],
    'multiply_str_int':
        ['a = "ERROR" * 5', ['incompatible_types'], []],
    'unary_and_sub_int_int':
        ['-(5)+0', ['incompatible_types'], []],
    'simple_unary_op':
        ['+1', ['incompatible_types'], []],
    'unary_compare':
        ['-1 < 5', ['incompatible_types'], []],
    'iadd_int_int':
        ['a=0\na+=5\na', ['incompatible_types', 'unused_variable', 'initialization_problem', 'overwritten_variable'],
         []],
    'iadd_str_int':
        ['a=""\na+=5\na', ['unused_variable', 'initialization_problem', 'overwritten_variable'],
         ['incompatible_types']],
    'iadd_undefined':
        ['a+=5\na', ['unused_variable', 'overwritten_variable'], ['initialization_problem']],
    'iadd_unread':
        ['a=0\na+=5', ['initialization_problem', 'overwritten_variable'], ['unused_variable']],
    'string_in_string':
        ['"" in ""', ['incompatible_types'], []],
    "string_in_tuple_str":
        ['"" in ("", "")', ['incompatible_types'], []],
    "string_in_int":
            ['"" in 5', [], ['incompatible_types']],
    "string_in_tuple_int":
        ['"" in (5, 4)', [], ['incompatible_types']],
    "string_var_in_tuple_strings":
        ['food = "fruitless"\nfood in ("apple", "banana")', ['incompatible_types'], []],
    "iter_string_var_in_tuple_string":
        ['x = [""]\nfor y in x:\n y in ("", "")', ['incompatible_types'], [] ],
    "iter_int_var_in_tuple_int":
        ['x = [5]\nfor y in x:\n y in (3,)', ['incompatible_types'], [] ],
    "iter_int_var_in_tuple_string":
        ['x = [4]\nfor y in x:\n y in ("", "")', [], ['incompatible_types'] ],

    # Lambda
    'lambda_add':
        ['a = lambda: 0\nb=a()\nb+5', ['incompatible_types'], []],

    # Handle function definitions
    'uncalled_function':
        ['def named(x):\n\tprint(x)\n', ['initialization_problem'], ['unused_variable']],
    'called_int_function':
        ['def int_func(x):\n\treturn 5\nint_func(10)', [], []],
    'called_constant_function':
        ['def x():\n    return 4\nx()', ['unused_variable'], []],
    'overwritten_parameter':
        ['def x(v):\n  v = 0\n  return v\nx()', [], ['overwritten_variable']],
    'overwritten_variable_in_function':
        ['def x():\n  v = 0\n  v = 7\n  return v\nx()', [], ['overwritten_variable']],
    # Actions after returning
    'return_after_return':
        ['def x():\n    return 5\n    return 4\nx()', [], ['action_after_return']],
    'action_after_return_on_both_branches':
        ['def x():\n  if True:\n    return 4\n  else:\n    return 3\n  a = 0\n  print(a)\nx()', [],
         ['action_after_return']],
    # Function with subtypes
    'function_with_subtypes_add_int_list_int':
        ['def add_first(a_list):\n    for element in a_list:\n        return element + 5\nprint(add_first([1]))',
         ['incompatible_types'], []],
    'function_with_subtypes_add_int_list_str':
        ['def add_first(a_list):\n    for element in a_list:\n        return element + 5\nprint(add_first(["1"]))', [],
         ['incompatible_types']],
    'function_with_subtypes_add_int_primitive_int':
        ['def add_first(a_list):\n    for element in a_list:\n        return element + 5\nprint(add_first(1))', [],
         ['iterating_over_non_list']],
    'function_with_subtypes_add_int_primitive_str':
        ['def add_first(a_list):\n    for element in a_list:\n        return element + 5\nprint(add_first("1"))', [],
         ['incompatible_types']],
    # Out of scope
    'read_out_of_scope':
        ['def x(parameter):\n    return parameter\nx(0)\nparameter', [], ['read_out_of_scope']],
    'read_inside_of_scope':
        ['def x(parameter):\n    return parameter\nx(0)', ['read_out_of_scope'], []],
    'read_not_out_of_scope':
        ['def x():\n    if 1:\n        y=0\n    else:\n        y=1\n    y\nx()\nx()', ['read_out_of_scope'], []],

    # Function with annotations
    'annotated_parameters_correct':
        ['def x(n: int):\n    1+n\nx(5)', ['parameter_type_mismatch', 'incompatible_types'], []],
    'annotated_parameters_wrong_argument':
        ['def x(n: int):\n    1+n\nx("5")', [], ['parameter_type_mismatch']],
    'annotated_parameters_wrong_parameter':
        ['def x(n: int):\n    "1"+n\nx(5)', [], ['incompatible_types']],
    'annotated_parameters_append_list':
        ['def x(l: list):\n    l.append(1)\nx([])', ['parameter_type_mismatch', 'incompatible_types'], []],
    'annotated_returns_int':
        ['def x()->int:\n    return "Wrong"\nx()', [], ['multiple_return_types']],
    'annotated_returns_list_int':
        ['def x()->[int]:\n    return [1,2,3]\nx().append(4)', ['multiple_return_types', 'incompatible_types'], []],

    'append_to_empty_list':
        ['a = []\na.append(1)\nprint(a)', ['initialization_problem', 'unused_variable'], []],
    'append_to_non_empty_list':
        ['a = [1]\na.append(1)\nprint(a)', ['initialization_problem', 'unused_variable'], []],
    'append_to_undefined':
        ['a.append(1)\nprint(a)', ['unused_variable'], ['initialization_problem']],
    'append_to_unread':
        ['a=[]\na.append(1)', ['initialization_problem'], ['unused_variable']],
    'append_to_number':
        ['a=1\na.append(1)\nprint(a)', [], ['append_to_non_list']],

    'append_and_index':
        ['x=[]\nx.append(1)\nx[0]+1', ['incompatible_types'], []],
    'indexing_used':
        ['mag1 = 0\nmercalli = [0]\nmag1 = mercalli[mag1]\nmag1',
         ['initialization_problem', 'unused_variable', 'overwritten_variable'], []],

    'created_list_but_unread':
        ['old = [1,2,3]\nnew=[]\nfor x in old:\n\tnew.append(x)', [], ['unused_variable']],
    'created_list_but_undefined':
        ['old = [1,2,3]\nfor x in old:\n\tnew.append(x)\nprint(new)', [], ['initialization_problem']],

    'builtin_float_converter':
        ['a = float(5)\nb = "test"\nprint(a+b)', [], ['incompatible_types']],

    # Double iteration
    'iterate_over_list_of_tuples':
        ['for x,y in [(1,2), (3,4)]:\n    x, y', ['initialization_problem'], []],
    'iterate_over_items':
        ['record = {"A": 5, "B": 6}\nfor x,y in record.items():\n    x, y', ['initialization_problem'], []],
    'iterate_over_items_add':
        ['record = {"A": 5, "B": 6}\nfor x,y in record.items():\n    x+"", y+0',
         ['initialization_problem', "incompatible_types"], []],

    # Tuple, Multiple Assignment
    'multiple_assignment': ['a,b = 1,2\n1+a\nb', ['incompatible_types'], []],
    'tuple_index': ['tuple_box = (6, 8, 4)\nprint(tuple_box[0])', [], []],

    # Sets
    'set_creation': ['a = set([1,2,3])\nprint(a)', ['initialization_problem'], []],
    'set_addition': ['a = {"dog"}\na+"cat"', [], ['incompatible_types']],
    'set_union_good': ['a = {"dog"}\na|{"cat"}', ['incompatible_types'], []],
    'set_union_bad': ['a = {"dog"}\na|"cat"', [], ['incompatible_types']],

    # Dictionaries
    'set_key_in_dict': ['a = {}\na[1] = 0', [], []],
    'use_key_in_dict': ['a = {"x": 5, "y": "T"}\na["x"]+5', ['incompatible_types'], []],
    'use_key_in_lod': ['x=[{"a": 0, "b": True}, {"a": 1, "b": False}]\ny=x[0]\nz=y["a"]+0', ['incompatible_types'], []],
    'use_chained_key_in_lod': ['x=[{"a": 0, "b": True}, {"a": 1, "b": False}]\nnot x[1]["b"]', ['incompatible_types'],
                               []],
    'iterate_over_lod': ['ls=[{"a": 0, "b": True}, {"a": 1, "b": False}]\nfor x in ls:\n    x["a"]+0',
                         ['incompatible_types'], []],
    # TODO: Add stronger assertion this one, it shouldn't be a good one
    'incorrect_dict_iteration': ['dict = {"T": "V"}\nfor key,value in dict:\n    print(key)', [], []],
    'incorrect_dict_iteration2': ['dict = {"T": 0}\nfor i in dict:\n print(i, dict[i])', [], []],

    # Record type
    'record_type_simple': [
        ('Dog = {"Name": str, "Age": int, "Fluffy": bool}\n'
         'def do_stuff(a_dog: Dog) -> Dog:\n'
         '    a_dog["Name"]+""\n'
         '    a_dog["Age"] = a_dog["Age"]+5\n'
         '    return a_dog\n'
         'ada = {"Name": "Ada", "Age": 2, "Fluffy": True}\n'
         'do_stuff(ada)["Name"] + ""'
         ), ['incompatible_types'], []
    ],
    'dict_with_setter': [
        ('person1={"Name": "Babbage", "Age": 3}\n'
            'def make_older(a_dog: dict) -> int:\n'
            '    a_dog["Age"]+=1\n'
            '    return a_dog["Age"]\n'
            'make_older(person1) + 0'), ['incompatible_types', 'type_changes'], []
    ],
    'function_returns_typed_dict': [
        ('def count_words(words: str) -> {str: int}:\n'
         '  counts = {}\n'
         '  for word in words.split(","):\n'
         '    if word not in counts:\n'
         '      counts[word] = 0\n'
         '    counts[word] += 1\n'
         '  return counts\n'
         'count_words("alpha,alpha,beta,alpha")'), ['multiple_return_types'], []
    ],
    'record_type_in_list': [
        ('Test = {"a": str}\n'
         'def x(y: [Test]):\n'
         '  return y[0]["a"]\n'
         'x([{"a": "apple"}])'),
         ['Parameter Type Mismatch'], []
    ],

    # While
    'while_until_input': [
        'user = input("Give a word.")\nwhile user:\n    print(user)\n    user = input("Give another word.")',
        ['unused_variable'], []],

    # With
    'with_open':
        ['with open("A") as a:\n    print(a)', ['initialization_problem'], []],

    # List comprehensions
    'list_comprehension':
        ['a = [5 for x in range(100)]\nfor i in a:\n    5+i', ['iterating_over_non_list', 'incompatible_types'], []],

    # return_outside_function
    'no_return_outside_function':
        ['def x():\n    return 5\nx()', ['return_outside_function'], []],
    'return_outside_function':
        ['def x():\n    pass\nreturn 5\nx()', [], ['return_outside_function']],

    # Classes
    'class_definition':
        [
            'class A:\n    y = 0\n    def __init__(self, x):\n        self.x = 0\n        self.test()\n    def test(self):\n        self.x = 5\nA()',
            [], []],
    'instance_assignment':
        ['class A:\n pass\na = A()\na.b = 0\nb', [], ['initialization_problem']],
    'instance_assignment_alt':
        ['class A:\n pass\na = A()\na.b = 0\na.b', ['initialization_problem'], []],
    'constructor_assignment':
        ['class A:\n def __init__(self):\n  self.x=0\na=A()\na.x+""', [], ['incompatible_types']],
    'parameterized_constructor_assignment':
        [dedent("""
                class A:
                    def __init__(self, f):
                        self.y = f
                a = A(7)
                a.y + ''"""), [], ['incompatible_types']],
    'complex_destructuring':
        [dedent("""
                class Player:
                    def __init__(self):
                        self.health = 100
                class World:
                    def __init__(self):
                        self.p = Player()
                w = World()
                w.p.health, w.p = (5, Player())
                """), ['incompatible_types'], []],
    'class_type_promotion':
        [dedent("""
                class Enemy:
                    def __init__(self):
                        self.health = 0
                class Player:
                    def __init__(self):
                        self.enemies = []
                class World:
                    def __init__(self):
                        self.p = Player()
                w = World()
                w.p.enemies.append(Enemy())
                w.p.enemies[0].health + 100
                """), ['incompatible_types'], []],

    # Mutable Types
    'mutable_list_in_function':
        ['def t():\n    x = []\n    x.append(1)\n    return x\nfor x in t():\n    x + 1', ['incompatible_types'], []],
    # Importing
    'import_matplotlib':
        ['import matplotlib.pyplot as plt\nplt.hist([1,2,3])\nplt.show()', ['initialization_problem'], []],
    'import_random':
        ['from random import randint\na=randint(1,2)\n1+a', ['initialization_problem', 'incompatible_types'], []],

    'import_state_demographics':
        [
            "import state_demographics\n\n\nincome_list = state_demographics.get(\"Per Capita Income\",\"(None)\",'')\nfilter_income = []\nfor income in income_list:\n    if income > 28000:\n        income_list.append(filter_income)\nprint(filter_income)\n",
            [], []],
    'import_state_demographics2':
        [
            "import state_demographics\n\n\nincome_list = state_demographics.get(\"Per Capita Income\",\"(None)\",'')\nnew_income_list = 0\nfor income in income_list:\n    if income > 28000:\n        new_income_list = new_income_list + 1\nprint(new_income_list)\n",
            [], []],
    'filter_pattern':
        ['l = []\nfor x in l:\n    if x > 0:\n        x', [], []],
    'append_iter_var_to_list':
        ['x = []\nx.append(x)\nx', [], []],
    'append_multiple_types':
        ['x = []\nx.append(7)\nx.append("F")\nx', [], ['type_change_append']],
    'bad_append_iter_var_to_list':
        ['x = []\nx.append(x)\nx.append(7)\nx', [], ['type_change_append']],
    'function_with_parameter':
        ['def x(y):\n    y\nx()', [], []],
    'function_returns_None':
        ['def x():\n    return\nx()', [], []],
    'mutually_recursive_call':
        ['def y():\n    x()\ndef x():\n    y()\nx()', [], ['recursive_call']],
    'recursive_call':
        ['def x():\n    x()\nx()', [], ['recursive_call']],
    'overwritten_double_nested_branch':
        ['b= 0\nif True:\n    if True:\n        b=0\nb', ['possible_initialization_problem'], []],
    # Overwritten in one branches
    'overwritten_in_one_branch_alt':
        ['a = 0\nif True:\n\ta = 1\na', ['possible_initialization_problem'], []],
    'filter_pattern2':
        ["t = 0\nfor x in []:\n    if x:\n        t = t + 1\nprint(t)", ['possible_initialization_problem'], []],
    'read_out_scope2':
        ["x = ''\ndef y():\n    return x\ny()", ['unused_variable'], []],

    'read_out_scope_double_branch':
        ["def x():\n  if True:\n    y=0\n  else:\n    y=1\n  y\nx()",
         ['unused_variable', 'read_out_of_scope'], []],

    # Calling functions from within functions
    'call_function_within_function':
        ['def z():\n     return b\ndef y():\n    b = 0\n    z()\n    return b\ndef x():\n    y()\nx()',
         ['unused_variable'], ['read_out_of_scope']],

    # While loop with possibly unused body
    'while_body_and_conditional_uses_variable':
        ['a = 10\nwhile a:\n    a -= 1', ['unused_variable'], []],
    'while_body_uses_variable':
        ['a = 10\nwhile True:\n    a -= 1', [], ['unused_variable']],
    'while_body_possibly_defines_variable':
        ['while True:\n    a=0\na', [], ['possible_initialization_problem']],
    'while_body_defined_variable':
        ['a=0\nwhile True:\n    a=0\na', ['possible_initialization_problem'], []],

    # Generators
    'add_list_to_list_comprehension':
        ['[1]+[a for a in [1,2,3]]', ['unused_variable', "incompatible_types"], []],
    'add_set_to_set_comprehension':
        ['{4}+{a for a in [1,2,3]}', ['unused_variable'], ["incompatible_types"]],
    'union_set_to_set_comprehension':
        ['{4}|{a for a in [1,2,3]}', ['unused_variable', "incompatible_types"], []],
    'int_membership_in_dictionary':
        ['3 in {a:a for a in [1,2,3]}', ['unused_variable', "incompatible_types"], []],
    'int_membership_in_comprehension':
        ['4 in (a for a in [1,2,3])', ['unused_variable', "incompatible_types"], []],

    'prevent_empty_iteration_in_appended_list':
        ['eles = [1,2,3]\nx = []\nfor ele in eles:\n    x.append(ele)\nfor e2 in x:\n    e2+1',
         ['iterating_over_empty_list'], []],

    'prevent_empty_iteration_dict':
        ['x={"A":5}\nfor y in x:\n y', ['iterating_over_empty_list'], []],
    
    # Slices
    'function_call_in_slice':
        ['def x(): return 2\n"TEST"[:x()]', ['unused_variable'], []],

    # Built-in modules
    'import_string_letters':
        ['import string\nstring.letters+""', ['incompatible_types'], []],

    # Nested function definitions
    'function_nested_inside_if':
        ['if False:\n  def x():\n    pass', [], ['nested_function_definition']],
    'function_toplevel':
        ['def x():\n  if False:\n    pass', ['nested_function_definition'], []],
    'no_function_whatsoever':
        ['if False:\n  pass', ['nested_function_definition'], []],
    'unused_result_function':
        ['abs(-5)', [], ['unused_returned_value']],
    'unused_result_method':
        ['"".strip()', [], ['unused_returned_value']],
    'unused_result_custom_function':
        ['def x(): return 7\nx()', [], ['unused_returned_value']],
    'used_result_custom_function':
        ['def x(): return 7\nprint(x())', ['unused_returned_value'], []],
    'used_result_function':
        ['print(abs(-5))', ['unused_returned_value'], []],

    # Reused variable in outer scope
    'function_reused_outer_variable_write':
        ['def y(b): return b+"!"\ndef x(a): return y(a)+"?"\na="Hi"\nx(a)',
         ['write_out_of_scope'], []],

    # Dictionary stuff
    'dictionary_literal_key_access_int':
        ['person = {"Name": "Charles", "Age": 4}\nperson["Age"] + 1',
         ['incompatible_types'], []],
    'dictionary_literal_key_access_int_bad':
        ['person = {"Name": "Charles", "Age": 4}\nperson["Age"] + "1"',
         [], ['incompatible_types']],
    'dictionary_literal_key_access_str':
        ['person = {"Name": "Charles", "Age": 4}\nperson["Name"] + 1',
         [], ['incompatible_types']],
    'dictionary_literal_key_access_str_bad':
        ['person = {"Name": "Charles", "Age": 4}\nperson["Name"] + "1"',
         ['incompatible_types'], []],

    # Enumerate
    'enumerate_function_in_function':
        ['def do_it(data):\n    v = 0\n    for i, _ in enumerate(data):\n        v += i\n    return v\ndo_it("ab")',
         ['unused_variable'], []]
}


class TestCode(unittest.TestCase):
    pass


SILENCE_EXCEPT = 'enumerate_function_in_function' # 'possibly_overwritten_in_elif_branch' # None  # 'read_not_out_of_scope'


def make_tester(internal_name, code, nones, somes):
    def test_code(self):
        contextualize_report(code)
        tifa = pedal.tifa.Tifa()
        try:
            result = tifa.process_code(code)
        except Exception as e:
            raise type(e)(str(e) +
                          ' in code:\n%s' % code)
        if not result.success:
            traceback.print_tb(result.error.__traceback__)
            #self.fail("Error message in\n" + code + "\n" +
            #          str(result.error))
            self.fail(f"Error occurred: {str(result.error)}\nName: {internal_name}\n----\n{code}\n----\n")
        for none in nones:
            if result.issues.get(none, []):
                print("")
                pprint(result.variables)
                self.fail(f"Incorrectly detected `{none}`\nName: {internal_name}\n----\n{code}\n----\n")
        for some in somes:
            if not result.issues.get(some, []):
                self.fail(f"Failed to detect `{some}`\nName: {internal_name}\n----\n{code}\n----\n")
    test_code.__name__ = "test_code_"+str(internal_name)
    return test_code


for name, (code, nones, somes) in unit_tests.items():
    if SILENCE_EXCEPT in (None, "None") or SILENCE_EXCEPT == name:
        setattr(TestCode, 'test_code_{}'.format(name), make_tester(name, code, nones, somes))


class TestVariables(unittest.TestCase):
    def test_type_comparisons(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('my_int=0\nmy_str="A"\nmy_list=[1,2,3]')
        variables = result.top_level_variables
        self.assertEqual(len(variables), 3)
        # Integer variable
        my_int = variables['my_int'].type
        self.assertTrue(my_int.is_equal('num'))
        self.assertTrue(my_int.is_equal('NumType'))
        self.assertTrue(my_int.is_equal(int))
        self.assertFalse(my_int.is_equal(float))
        self.assertFalse(my_int.is_equal(str))
        # String variable
        my_str = variables['my_str'].type
        self.assertTrue(my_str.is_equal('str'))
        self.assertTrue(my_str.is_equal('StrType'))
        self.assertTrue(my_str.is_equal(str))
        # List variable
        my_list = variables['my_list'].type
        self.assertTrue(my_list.is_equal('list'))
        self.assertTrue(my_list.is_equal('ListType'))
        self.assertTrue(my_list.is_equal(list))
        # List subtype
        self.assertTrue(my_list.element_type.is_equal(int))
        self.assertTrue(my_list.element_type.is_equal('num'))
        self.assertFalse(my_list.element_type.is_equal(float))
        self.assertTrue(my_list.element_type.is_equal('NumType'))

    def test_variable_def_use(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('a=0\nb=0\nb\nc')
        issues = result.issues
        # Unused variables
        self.assertEqual(len(issues['unused_variable']), 1)
        unused_variables = [i.fields['name'] for i in issues['unused_variable']]
        self.assertIn('a', unused_variables)
        self.assertNotIn('b', unused_variables)
        self.assertNotIn('c', unused_variables)
        # Uninitalized variables
        self.assertEqual(len(issues['initialization_problem']), 1)
        unset_variables = [i.fields['name'] for i in issues['initialization_problem']]
        self.assertNotIn('a', unset_variables)
        self.assertNotIn('b', unset_variables)
        self.assertIn('c', unset_variables)

    def test_weirdness_used_function(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('def a(x):\n    return x\na(1)')
        self.assertTrue(result.success)

        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('def a(x):\n    return x\n')
        issues = result.issues
        self.assertEqual(issues['unused_variable'][0].message,
                         "The function a was "
                         "given a definition on line 1, but was never used "
                         "after that.")

    def test_bad_parameter_found(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('def a(x: int):\n    return x\na("hello")')
        issues = result.issues
        self.assertEqual(issues['parameter_type_mismatch'][0].message,
                         "You defined the parameter x as an integer. However, the argument passed to that parameter on line 3 was a string. The formal parameter type must match the argument's type.")

    def test_bad_parameter_found_complex(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('def a(x: list[int]):\n    return x\na(5)')
        issues = result.issues
        self.assertEqual(issues['parameter_type_mismatch'][0].message,
                         "You defined the parameter x as a list of integers. However, the argument passed to that parameter on line 3 was an integer. The formal parameter type must match the argument's type.")

    def test_weirdness_tate_import_example(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(dedent('''
        import tate
        art = tate.get_artwork()
        h = [a['dimensions']['height'] for a in art if a['artist']['birth']['year'] > 1000]
        b = [a['artist']['birth']['year'] for a in art 
             if a['artist']['birth']['year'] > 1000]
        import matplotlib.pyplot as plt
        plt.scatter(b, h, alpha=.1)
        plt.show()'''))
        self.assertTrue(result.success)

    '''
    def test_tifa_graceful_errors(self):
        student_code = '1 + "Banana"'
        set_source(student_code)
        exception = commands.run_student()
        self.assertIsNotNone(exception)'''

    def test_json_types(self):
        t = normalize.get_pedal_type_from_json({'type': 'NumType'})
        self.assertIsInstance(t, defs.NumType)
        t = normalize.get_pedal_type_from_json({'type': 'StrType'})
        self.assertIsInstance(t, defs.StrType)
        t = normalize.get_pedal_type_from_json({'type': 'BoolType'})
        self.assertIsInstance(t, defs.BoolType)
        t = normalize.get_pedal_type_from_json({'type': 'NoneType'})
        self.assertIsInstance(t, defs.NoneType)
        t = normalize.get_pedal_type_from_json({'type': 'ListType',
                                 'subtype': {'type': 'NumType'}})
        self.assertIsInstance(t, defs.ListType)
        self.assertIsInstance(t.element_type, defs.NumType)

        l1 = {'type': 'LiteralStr', 'value': 'First'}
        l2 = {'type': 'LiteralStr', 'value': 'Second'}
        v1, v2 = {'type': 'StrType'}, {'type': 'NumType'}
        t = normalize.get_pedal_type_from_json({'type': 'DictType', 'literals': [l1, l2],
                                 'values': [v1, v2]})
        self.assertIsInstance(t, defs.DictType)
        self.assertIsInstance(t.element_types, list)
        self.assertEqual(len(t.element_types), 2)
        self.assertIsInstance(t.element_types[0][0], defs.LiteralStr)
        self.assertEqual(t.element_types[0][0].value, 'First')
        self.assertIsInstance(t.element_types[1][0], defs.LiteralStr)
        self.assertEqual(t.element_types[1][0].value, 'Second')
        self.assertIsInstance(t.element_types, list)
        self.assertEqual(len(t.element_types), 2)
        self.assertIsInstance(t.element_types[0][1], defs.StrType)
        self.assertIsInstance(t.element_types[1][1], defs.NumType)

        # Try to parse the Broadway type definition
        complex_type = {"type": "ModuleType",
                        "fields": {
                            'get': {"type": "ListType", "empty": False,
                                    "subtype": {"type": "NumType"}},

                            'get_shows':
                                {"type": "ListType", "subtype":
                                    {"type": "DictType", "literals": [{"type": "LiteralStr", "value": 'Statistics'},
                                                                      {"type": "LiteralStr", "value": 'Show'},
                                                                      {"type": "LiteralStr", "value": 'Date'}],
                                     "values": [
                                         {"type": "DictType", "literals": [{"type": "LiteralStr", "value": 'Capacity'},
                                                                           {"type": "LiteralStr",
                                                                            "value": 'Attendance'},
                                                                           {"type": "LiteralStr",
                                                                            "value": 'Performances'},
                                                                           {"type": "LiteralStr", "value": 'Gross'},
                                                                           {"type": "LiteralStr",
                                                                            "value": 'Gross Potential'}], "values": [
                                             {"type": "NumType"},
                                             {"type": "NumType"},
                                             {"type": "NumType"},
                                             {"type": "NumType"},
                                             {"type": "NumType"}]},
                                         {"type": "DictType", "literals": [{"type": "LiteralStr", "value": 'Type'},
                                                                           {"type": "LiteralStr", "value": 'Theatre'},
                                                                           {"type": "LiteralStr", "value": 'Name'}],
                                          "values": [
                                              {"type": "StrType"},
                                              {"type": "StrType"},
                                              {"type": "StrType"}]},
                                         {"type": "DictType", "literals": [{"type": "LiteralStr", "value": 'Day'},
                                                                           {"type": "LiteralStr", "value": 'Month'},
                                                                           {"type": "LiteralStr", "value": 'Year'},
                                                                           {"type": "LiteralStr", "value": 'Full'}],
                                          "values": [
                                              {"type": "NumType"},
                                              {"type": "NumType"},
                                              {"type": "NumType"},
                                              {"type": "StrType"}]}]}},

                        }}
        normalize.get_pedal_type_from_json(complex_type)

    def test_custom_module_import_weather(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('import weather\n' +
                          'rs = weather.get_weather()\n' +
                          'for r in rs:\n' +
                          '  0+r["Data"]["Precipitation"]',
                                   filename='student.py')
        self.assertTrue(result.success)
        self.assertNotIn('module_not_found', result.issues)

    def test_custom_module_import_police(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('import police_shootings\n'
                          'rs = police_shootings.get_shootings()\n'
                          'for r in rs:\n'
                          '  x=0+r["Person.Age"]\n'
                          'x', filename='student.py')
        self.assertTrue(defs.is_subtype(result.top_level_variables['x'].type, defs.NumType()))
        self.assertTrue(result.success)
        self.assertNotIn('module_not_found', result.issues)

    def test_custom_module_import_babbages(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('import babbages\n'
                          'print(babbages)', filename='student.py')
        self.assertIn('module_not_found', result.issues)

    def test_custom_module_import_broadway(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(dedent("""
            import broadway
            report_list = broadway.get_shows()
            total= 0
            for broadway in report_list:
                if broadway ["Show"]["Type"] == "Musical":
                    total = total + 1
            print(total)"""), filename='student.py')
        self.assertTrue(result.success)
        self.assertNotIn('module_not_found', result.issues)

    def test_get_types(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('a=0\na="Hello"\na=[1,2,3]\na=1')
        a = result.top_level_variables['a']
        self.assertTrue(a.was_type('num'))
        self.assertTrue(a.was_type(int))
        self.assertTrue(a.was_type('NumType'))
        self.assertTrue(a.was_type(defs.NumType))
        self.assertFalse(a.was_type(bool))
        self.assertFalse(a.was_type(defs.DictType))
        self.assertTrue(a.was_type('ListType'))

        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('credits=[1,2,3,4]\nfor credit in credits:\n  print(credit)')
        credit = result.top_level_variables['credit']
        self.assertFalse(credit.was_type(list))
        self.assertTrue(credit.was_type(int))

    def test_dict_iteration(self):
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code('dict = {"T": 0}\nfor i in dict:\n    print(i, dict[i])')
        self.assertTrue(result.success)

    def test_classes(self):
        program = 'class A:\n def __init__(self):\n  self.x=0\na=A()\na.x+""'
        # print(indent(program, '    '))
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program)
        # self.assertTrue(tifa.report['tifa']['success'])
        # pprint(tifa.report['tifa'])
        # print(tifa.report['tifa']['top_level_variables']['a'].type.parent.fields)
        self.assertFalse(result.error)

        program = dedent("""
                class Enemy:
                    def __init__(self):
                        self.health = 0
                class Player:
                    def __init__(self):
                        self.enemies = []
                class World:
                    def __init__(self):
                        self.p = Player()
                w = World()
                w.p.enemies.append(Enemy())
                w.p.enemies[0].health + 100
                """)
        tifa = pedal.tifa.Tifa()
        tifa.process_code(program)
        # pprint(tifa.report['tifa'])
    
    def test_locations(self):
        program = dedent("""
                a = 0
                "Ignore"
                "This"
                "Line"
                def unnecessary():
                    pass
                unnecessary
                """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program)
        self.assertEqual(result.issues['unused_variable'][0].location.line, 2)


    def test_weird_indexing_behavior(self):
        program = dedent("""
        movie_list = [{'Title':'Iron Man', 'Release Date':'May 2, 2008','Budget':'140','Running Time':'126'},
              {'Title':'Iron Man 2', 'Release Date':'May 7, 2010','Budget':'200', 'Running Time':'125'},
              {'Title':'Iron Man 3', 'Release Date':'May 3, 2013','Budget':'200', 'Running Time':'130'}]
        total=0
        for title in movie_list:
            total=total + 'title'['Running Time']
        print(total)
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program)
        #self.assertNotIn('incompatible_types', result.issues)
        self.assertIn('invalid_indexing', result.issues)

    def test_not_finding_unused_return_value(self):
        program = dedent("""
            #import classics
            import matplotlib.pyplot as plt
            book_downloads = []
            #report_list = classics.get_books(test=True)
            report_list = [{'bibliography': {'type': 'Text'},
                            'metadata': {'downloads': 0}}]
            for report in report_list:
                if report["bibliography"]["type"] == "Text":
                    book_downloads.append(report["metadata"]["downloads"])
            plt.hist(book_downloads)
            plt.ylabel("Number of Books")
            plt.xlabel("downloads")
            plt.title("Spreads of Book Downloads")
            plt.show()
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertNotIn('unused_returned_value', result.issues)
        self.assertFalse(result.issues)

    @unittest.skip("Currently broken, need to increase sophistication of typeddict support")
    def test_typedict_book(self):
        program = dedent("""
            from cisc108 import *
            class Book(TypedDict):
                title: str
                pages: int
                hardcover: bool
            
            def page_count(a_book: Book) -> int:
                return a_book['pages']
                
            hp = {'title': 'Harry Potter',
                  'pages': 300,
                  'hardcover': false}
            
            assert_equal(page_count(hp), 300)
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertFalse(result.issues)

    def test_unnecessary_second_branch(self):
        program = dedent("""# Remember to initialize the variables!
age = 20
has_license = True
# Keep this If/else structure unchanged, but
# fill in the blanks with the code above
if has_license == True:
    if age >= 21:
        print("Can drink")
        if age < 1000:
            pass
        else:
            print("Too old")
            
    else:
        print("Too young")
        
else:
    print("Doesn't have a license")""")
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertTrue(result.issues)

    def test_complex_record_type(self):
        program = dedent("""
Test = {"a": str}

def x(y: [Test]):
    return y[0]['a']

print(x([{"a": "apple"}]))""")
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertFalse(result.issues)

    def test_new_class_record_type(self):
        program = dedent("""
        from dataclasses import dataclass
        
        @dataclass
        class Dog:
            name: str
            age: int
            fuzzy: bool
            
        def is_fuzzy(a_dog: Dog) -> bool:
            return a_dog.fuzzy
            
        ada = Dog('Ada Bart', 4, True)
        print(ada.name)
        
        print(is_fuzzy(ada))
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertFalse(result.issues)

    def test_new_class_record_type_bad_instance(self):
        program = dedent("""
        class Dog(record):
            name: str
            age: int
            fuzzy: bool

        def is_fuzzy(a_dog: Dog) -> bool:
            return a_dog.fuzzy

        ada = Dog('Ada Bart', 4)
        print(ada.name)

        print(is_fuzzy(ada))
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertTrue(result.issues)

    def test_dataclass_mistake(self):
        program = dedent("""
        from dataclasses import dataclass
        
        @dataclass
        class Dog:
            name: str
            age: int
            fuzzy: bool
            
        @dataclass
        class Cat:
            name: str
            age: int
            fuzzy: bool

        def is_fuzzy(a_dog: Dog) -> bool:
            return a_dog.fuzzy

        ada = Dog('Ada Bart', 4, True)
        print(ada.name)

        print(is_fuzzy(ada))
        
        adacat = Cat("Ada Bart", 4, True)
        print(is_fuzzy(adacat))
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertIsNone(result.error)
        self.assertTrue(result.issues)

    def test_dataclass_mistake_lists(self):
        program = dedent("""
        from dataclasses import dataclass

        @dataclass
        class Dog:
            name: str
            age: int
            fuzzy: bool

        @dataclass
        class Cat:
            name: str
            age: int
            fuzzy: bool

        def all_fuzzy(dogs: list[Dog]) -> bool:
            return all(a_dog.fuzzy for a_dog in dogs)

        ada = Dog('Ada Bart', 4, True)
        print(ada.name)

        print(all_fuzzy([ada]))

        adacat = Cat("Ada Bart", 4, True)
        print(all_fuzzy([adacat]))
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertIsNone(result.error)
        self.assertTrue(result.issues)
        self.assertEqual("You defined the parameter dogs as a list of Dog. "
                         "However, the argument passed to that parameter on line 25 was a list of Cat. "
                         "The formal parameter type must match the argument's type.",
                         result.issues['parameter_type_mismatch'][0].message)

    def test_dataclass_bad_constructor_call_order(self):
        program = dedent("""
        from dataclasses import dataclass

        @dataclass
        class Dog:
            name: str
            age: int
            fuzzy: bool
        
        ada = Dog('Ada Bart', True, 4)
        print(ada.name)
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertIsNone(result.error)
        self.assertTrue(result.issues)
        self.assertEqual("You defined the field age as an integer. However, the argument passed to that field's "
                         "parameter in the constructor function on line 10 was a boolean. The formal field type "
                         "must match the argument's type.",
                         result.issues['field_type_mismatch'][0].message)

    def test_dataclass_bad_constructor_call_arity(self):
        program = dedent("""
        from dataclasses import dataclass

        @dataclass
        class Dog:
            name: str
            age: int
            fuzzy: bool

        ada = Dog('Ada Bart', 4)
        print(ada.name)
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertIsNone(result.error)
        self.assertTrue(result.issues)
        self.assertEqual("The constructor function Dog was given the wrong number of arguments. You should have "
                         "had 3 arguments, but instead you had 2 arguments.",
                         result.issues['incorrect_arity'][0].message)

    def test_dataclass_bad_constructor_call_extra(self):
        program = dedent("""
        from dataclasses import dataclass

        @dataclass
        class Dog:
            name: str
            age: int
            fuzzy: bool

        ada = Dog('Ada Bart', 4, True, "is cute!")
        print(ada.name)
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(program, filename="student.py")
        self.assertIsNone(result.error)
        self.assertTrue(result.issues)
        self.assertEqual("The constructor function Dog was given the wrong number of arguments. You should have "
                         "had 3 arguments, but instead you had 4 arguments.",
                         result.issues['incorrect_arity'][0].message)

    def test_import_second_file(self):
        main_program = dedent("""
        import monsters
        
        monsters.number + ""
        """)
        monster_file = dedent("""number = 7\nalpha = 'test'""")
        contextualize_report(Submission(files={'student.py': main_program, 'monsters.py': monster_file}))
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(main_program, filename="student.py")
        self.assertIsNone(result.error)
        self.assertTrue(result.issues)
        self.assertEqual("You used an addition operation with an integer and a string on line 4. But you can't do that with that operator. Make sure both sides of the operator are the right type.",
                         result.issues['incompatible_types'][0].message)
        self.assertEqual(len(result.issues.get('unused_variable', [])), 0)

    def test_import_second_file_successful(self):
        main_program = dedent("""
        from monsters import Monster, dracula

        dracula.name + ""
        print(Monster("Wolfman"))
        """)
        monster_file = dedent("""
        from dataclasses import dataclass
        @dataclass
        class Monster:
            name: str
        dracula = Monster('Dracula')""")
        contextualize_report(Submission(files={'student.py': main_program, 'monsters.py': monster_file}))
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(main_program, filename="student.py")
        self.assertIsNone(result.error)
        self.assertFalse(result.issues)

    def test_provide_import_type_definitions(self):
        main_program = dedent("""
        import monsters

        monsters.number + ""
        """)
        tifa = pedal.tifa.Tifa()
        tifa_provide_module_type('monsters', {'number': normalize.normalize_type('int')})
        result = tifa.process_code(main_program, filename="student.py")
        self.assertIsNone(result.error)
        self.assertTrue(result.issues)
        self.assertEqual(
            "You used an addition operation with an integer and a string on line 4. But you can't do that with that operator. Make sure both sides of the operator are the right type.",
            result.issues['incompatible_types'][0].message)

    @unittest.skip
    def test_provide_module_definitions(self):
        main_program = dedent("""
        from monsters import dracula

        dracula.age + ""
        """)
        tifa = pedal.tifa.Tifa()
        import tests.test_files.monsters as monsters
        tifa_provide_module_type('monsters', normalize.normalize_type(monsters))
        result = tifa.process_code(main_program, filename="student.py")
        if result.error is not None:
            raise result.error
        self.assertIsNone(result.error)
        self.assertTrue(result.issues)
        self.assertEqual(
            "You used an addition operation with an integer and a string on line 4. But you can't do that with that operator. Make sure both sides of the operator are the right type.",
            result.issues['incompatible_types'][0].message)

    def test_weird_return_control_flow(self):
        main_program = dedent("""
        from dataclasses import dataclass
        
        @dataclass
        class Media:
            name: str
            kind: str
            duration: int
        
        def longest(media: list[Media]) -> str:
            movies_only = [i for i in media if i.kind == 'movie']
            if not movies_only:
                return 'no movies'
            key = lambda movie: movie.duration
            return max(movies_only, key=key).name
        
        print(longest([Media("Snoopy", "movie", 5)]))
        print(longest([]))
        """)
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(main_program, filename="student.py")
        if result.error:
            raise result.error
        self.assertIsNone(result.error)
        self.assertFalse(result.issues)

    def test_weird_increment_set_list(self):
        main_program = dedent("""from dataclasses import dataclass

@dataclass
class Forecast:
    reports: list[int]
    
def rainfall(weather: list[Forecast]) -> list[int]:
    reports = []
    for a_weather in weather:
        reports += a_weather.reports
    return reports

def total_rainfall(forecast: list[Forecast]) -> int:
    total = 0
    reports = rainfall(forecast)
    for measurement in reports:
        total += measurement
    return total

forecast1 = [Forecast([1,2,3])]
5+total_rainfall(forecast1)""")
        tifa = pedal.tifa.Tifa()
        result = tifa.process_code(main_program, filename="student.py")
        if result.error:
            raise result.error
        self.assertIsNone(result.error)
        self.assertFalse(result.issues)

if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Runtime assertions.

TODO: assert_has_class
TODO: assertGraphType, assertGraphValues
TODO: assert_coverage
TODO: assert_ran (latest run produced no expections)


assert_equal
assert_not_equal
assert_less
assert_less_equal
assert_greater
assert_greater_equal
assert_in
assert_not_in
assert_contains_subset
assert_is
assert_is_not
assert_is_none
assert_is_dataclass
assert_is_not_dataclass
assert_true
assert_false
assert_length_equal
"""
import re

try:
    from dataclasses import _FIELDS
except Exception:
    _FIELDS = '__dataclass_fields__'

from pedal.assertions.feedbacks import (RuntimeAssertionFeedback,
                                        SandboxedValue, ExactValue,
                                        RuntimePrintingAssertionFeedback, AssertionFeedback, assert_group)
from pedal.assertions.functions import function_not_available, name_is_not_a_function
from pedal.core.report import MAIN_REPORT
from pedal.core.feedback import CompositeFeedbackFunction
from pedal.sandbox import Sandbox
from pedal.sandbox.commands import check_coverage, get_call_arguments, get_student_data, evaluate
from pedal.sandbox.result import share_sandbox_context, is_sandbox_result, unwrap_value
from pedal.types.normalize import normalize_type, get_pedal_type_from_value
from pedal.types.new_types import is_subtype
from pedal.utilities.comparisons import equality_test


def errors(*executions):
    for e in executions:
        if e.is_error:
            return True
    return False


class assert_equal(RuntimeAssertionFeedback):
    """
    Determine if the ``left`` and ``right`` values are equal.

    Args:
        exact_strings (bool): Whether to require that strings be exactly the
            same, for each character. If False (the default), then strings will
            be normalized (lowercased, trailing decimals chopped, punctuation
            removed, lines are flattened, and all characters are sorted).
        delta (float): When comparing floats, how close the values must be.
            If delta is None, then the default Delta will be used (.001).
    """
    DELTA = .001
    justification_template = ("Left ({left}) and right ({right}) were not equal",
                              "Left ({left}) and right ({right}) were equal")
    _expected_verb = "to be equal to"
    _aggregate_verb = "Expected"
    _inverse_operator = "!="

    def __init__(self, left, right, exact_strings=False, delta=DELTA, **kwargs):
        super().__init__(SandboxedValue(left), SandboxedValue(right),
                         exact_strings=exact_strings, delta=delta, **kwargs)

    def condition(self, left, right, exact_strings, delta):
        """ Tests if the left and right are equal """
        return errors(left, right) or not equality_test(left.value, right.value, exact_strings, delta)


class assert_not_equal(RuntimeAssertionFeedback):
    """
    Determine if the ``left`` and ``right`` values are not equal.

    Args:
        exact_strings (bool): Whether to require that strings be exactly the
            same, for each character. If False (the default), then strings will
            be normalized (lowercased, trailing decimals chopped, punctuation
            removed, lines are flattened, and all characters are sorted).
        delta (float): When comparing floats, how close the values must be.
            If delta is None, then the default Delta will be used (.001).
    """
    DELTA = .001
    justification = "Left and right were equal"
    _expected_verb = "to not be equal to"
    _aggregate_verb = "Expected Anything But"
    _inverse_operator = "=="

    def __init__(self, left, right, exact_strings=False, delta=DELTA, **kwargs):
        super().__init__(SandboxedValue(left), SandboxedValue(right),
                         exact_strings=exact_strings, delta=delta, **kwargs)

    def condition(self, left, right, exact_strings, delta):
        """ Tests if the left and right are not equal """
        return equality_test(left.value, right.value, exact_strings, delta)


class assert_less(RuntimeAssertionFeedback):
    """
    Determine if the ``left`` is less than the ``right``.
    """
    justification = "Left is not less than right"
    _expected_verb = "to be less than"
    _inverse_operator = ">="

    def __init__(self, left, right, **kwargs):
        super().__init__(SandboxedValue(left), SandboxedValue(right), **kwargs)

    def condition(self, left, right):
        """ Tests if the left is greater or equal """
        return left.value >= right.value


class assert_less_equal(RuntimeAssertionFeedback):
    """
    Determine if the ``left`` is less than or equal to the ``right``.
    """
    justification = "Left is not less than or equal to the right"
    _expected_verb = "to be less than or equal to"
    _inverse_operator = ">"

    def __init__(self, left, right, **kwargs):
        super().__init__(SandboxedValue(left), SandboxedValue(right), **kwargs)

    def condition(self, left, right):
        """ Tests if the left is greater than the right """
        return left.value > right.value


class assert_greater(RuntimeAssertionFeedback):
    """
    Determine if the ``left`` is greater than the ``right``.
    """
    justification = "Left is not greater than right"
    _expected_verb = "to be greater than"
    _inverse_operator = "<="

    def __init__(self, left, right, **kwargs):
        super().__init__(SandboxedValue(left), SandboxedValue(right), **kwargs)

    def condition(self, left, right):
        """ Tests if the left is less than or equal to the right """
        return left.value <= right.value


class assert_greater_equal(RuntimeAssertionFeedback):
    """
    Determine if the ``left`` is greater than or equal to the ``right``.
    """
    justification = "Left is not greater than or equal to the right"
    _expected_verb = "to be greater than or equal to"
    _inverse_operator = "<"

    def __init__(self, left, right, **kwargs):
        super().__init__(SandboxedValue(left), SandboxedValue(right), **kwargs)

    def condition(self, left, right):
        """ Tests if the left is less than the right """
        return left.value < right.value


class assert_in(RuntimeAssertionFeedback):
    """
    Determine if the ``needle`` is in the ``haystack``.
    """
    justification = "Needle not in haystack"
    _expected_verb = ("to be in", "to contain")
    _inverse_operator = "not in"

    def __init__(self, needle, haystack, **kwargs):
        super().__init__(SandboxedValue(needle), SandboxedValue(haystack), **kwargs)

    def condition(self, needle, haystack):
        """ Tests if the needle is not in the haystack """
        return needle.value not in haystack.value


class assert_not_in(RuntimeAssertionFeedback):
    """
    Determine if the ``needle`` is in the ``haystack``.
    """
    justification = "Needle is in haystack"
    _expected_verb = ("to not be in", "to not contain")
    _inverse_operator = "in"

    def __init__(self, needle, haystack, **kwargs):
        super().__init__(SandboxedValue(needle), SandboxedValue(haystack), **kwargs)

    def condition(self, needle, haystack):
        """ Tests if the needle is in the haystack """
        return needle.value in haystack.value


class assert_contains_subset(RuntimeAssertionFeedback):
    """
    Determine if the ``needles`` are in the ``haystack``.
    """
    justification = "Needles not in haystack"
    _expected_verb = ("to be in", "to contain")
    _inverse_operator = "not in"

    def __init__(self, needles, haystack, **kwargs):
        super().__init__(SandboxedValue(needles), SandboxedValue(haystack), **kwargs)

    def condition(self, needles, haystack):
        """ Tests if the needle is not in the haystack """
        return not all(needle in haystack.value for needle in needles.value)


class assert_not_contains_subset(RuntimeAssertionFeedback):
    """
    Determine if the ``needles`` are not in the ``haystack``.
    """
    justification = "Needles in haystack"
    _expected_verb = ("to not be in", "to not contain")
    _inverse_operator = "in"

    def __init__(self, needles, haystack, **kwargs):
        super().__init__(SandboxedValue(needles), SandboxedValue(haystack), **kwargs)

    def condition(self, needles, haystack):
        """ Tests if the needle is not in the haystack """
        return all(needle in haystack.value for needle in needles.value)


class assert_is(RuntimeAssertionFeedback):
    """
    Determine if the ``left`` and ``right`` values are identical.
    """
    justification = "Left is not identical to right"
    _expected_verb = "to be identical to"
    _inverse_operator = "is not"

    def __init__(self, left, right, **kwargs):
        super().__init__(SandboxedValue(left), SandboxedValue(right), **kwargs)

    def condition(self, left, right):
        """ Tests if the left and right are equal """
        left = left.value._actual_value if left.is_sandboxed else left.value
        right = right.value._actual_value if right.is_sandboxed else right.value
        return left is not right


class assert_is_not(RuntimeAssertionFeedback):
    """
    Determine if the ``left`` and ``right`` values are not identical.
    """
    justification = "Left is identical to right"
    _expected_verb = "to not be identical to"
    _inverse_operator = "is"

    def __init__(self, left, right, **kwargs):
        super().__init__(SandboxedValue(left), SandboxedValue(right), **kwargs)

    def condition(self, left, right):
        """ Tests if the left and right are equal """
        left = left.value._actual_value if left.is_sandboxed else left.value
        right = right.value._actual_value if right.is_sandboxed else right.value
        return left is right


class assert_is_none(RuntimeAssertionFeedback):
    """
    Determine if the ``value`` is ``None``.
    """
    justification = "Value is not None"
    _expected_verb = "to be"
    _inverse_operator = "is not"

    def __init__(self, value, **kwargs):
        super().__init__(SandboxedValue(value), ExactValue("None"), **kwargs)

    def condition(self, left, right):
        """ Tests if the left and right are equal """
        if left.is_sandboxed:
            return left.value._actual_value is not None
        return left.value is not None


class assert_is_not_none(RuntimeAssertionFeedback):
    """
    Determine if the ``value`` is not ``None``.
    """
    justification = "Value is None"
    _expected_verb = "to not be"
    _inverse_operator = "is"

    def __init__(self, value, **kwargs):
        super().__init__(SandboxedValue(value), ExactValue("None"), **kwargs)

    def condition(self, left, right):
        """ Tests if the left and right are equal """
        if left.is_sandboxed:
            return left.value._actual_value is None
        return left.value is None


class assert_is_dataclass(RuntimeAssertionFeedback):
    """
    Determine if the ``value`` is a dataclass.
    """
    justification = "Value does not evaluate to a dataclass"
    _expected_verb = "to evaluate to"
    _inverse_operator = "to not evaluate to"

    def __init__(self, value, **kwargs):
        super().__init__(SandboxedValue(value), ExactValue("a dataclass"), **kwargs)

    def condition(self, left, right):
        """ Tests if the left evaluates to true """
        return not hasattr(left.value, _FIELDS)


class assert_is_not_dataclass(RuntimeAssertionFeedback):
    """
    Determine if the ``value`` is not a dataclass.
    """
    justification = "Value does not evaluate to a dataclass"
    _expected_verb = "to not evaluate to"
    _inverse_operator = "to evaluate to"

    def __init__(self, value, **kwargs):
        super().__init__(SandboxedValue(value), ExactValue("a dataclass"), **kwargs)

    def condition(self, left, right):
        """ Tests if the left evaluates to true """
        return hasattr(left.value, _FIELDS)


class assert_true(RuntimeAssertionFeedback):
    """
    Determine if the ``value`` is true.
    """
    justification = "Value does not evaluate to true"
    _expected_verb = "evaluates to"
    _inverse_operator = "does not evaluate to"

    def __init__(self, value, **kwargs):
        super().__init__(SandboxedValue(value), ExactValue("a true value"), **kwargs)

    def condition(self, left, right):
        """ Tests if the left evaluates to true """
        return not bool(left.value)


class assert_false(RuntimeAssertionFeedback):
    """
    Determine if the ``value`` is false.
    """
    justification = "Value does not evaluate to false"
    _expected_verb = "evaluates to"
    _inverse_operator = "does not evaluate to"

    def __init__(self, value, **kwargs):
        super().__init__(SandboxedValue(value), ExactValue("a true value"), **kwargs)

    def condition(self, left, right):
        """ Tests if the left evaluates to true """
        return bool(left.value)


class assert_length_equal(RuntimeAssertionFeedback):
    """
    Determine if the ``sequence`` has the ``length``.
    """
    justification = "Sequence does not have length"
    _expected_verb = ("to have the length", "to be the length of")
    _inverse_operator = "did not have the length"

    def __init__(self, sequence, length, **kwargs):
        super().__init__(SandboxedValue(sequence), SandboxedValue(length), **kwargs)

    def condition(self, sequence, length):
        """ Tests if the needle is not in the haystack """
        return len(sequence.value) != length.value


class assert_length_not_equal(RuntimeAssertionFeedback):
    """
    Determine if the ``sequence`` does not have the ``length``.
    """
    justification = "Sequence has length"
    _expected_verb = ("to not have the length", "to not be the length of")
    _inverse_operator = "had the length"

    def __init__(self, sequence, length, **kwargs):
        super().__init__(SandboxedValue(sequence), SandboxedValue(length), **kwargs)

    def condition(self, sequence, length):
        """ Tests if the needle is not in the haystack """
        return len(sequence.value) == length.value


class assert_length_less(RuntimeAssertionFeedback):
    """
    Determine if the ``sequence`` has less than the ``length``.
    """
    justification = "Sequence length is less than"
    _expected_verb = ("to have length less than", "to be less than the length of")
    _inverse_operator = "did not have less than the length"

    def __init__(self, sequence, length, **kwargs):
        super().__init__(SandboxedValue(sequence), SandboxedValue(length), **kwargs)

    def condition(self, sequence, length):
        """ Tests if the needle is not in the haystack """
        return len(sequence.value) >= length.value


class assert_length_less_equal(RuntimeAssertionFeedback):
    """
    Determine if the ``sequence`` has less or equal than the ``length``.
    """
    justification = "Sequence length is less than or equal to"
    _expected_verb = ("to have length less than or equal to",
                      "to be less than or equal to the length of")
    _inverse_operator = "did not have less than or equal to the length"

    def __init__(self, sequence, length, **kwargs):
        super().__init__(SandboxedValue(sequence), SandboxedValue(length), **kwargs)

    def condition(self, sequence, length):
        """ Tests if the needle is not in the haystack """
        return len(sequence.value) > length.value


class assert_length_greater(RuntimeAssertionFeedback):
    """
    Determine if the ``sequence`` has greater than the ``length``.
    """
    justification = "Sequence length is greater than"
    _expected_verb = ("to have length greater than", "to be greater than the length of")
    _inverse_operator = "did not have greater than the length"

    def __init__(self, sequence, length, **kwargs):
        super().__init__(SandboxedValue(sequence), SandboxedValue(length), **kwargs)

    def condition(self, sequence, length):
        """ Tests if the needle is not in the haystack """
        return len(sequence.value) <= length.value


class assert_length_greater_equal(RuntimeAssertionFeedback):
    """
    Determine if the ``sequence`` has greater than or equal to the ``length``.
    """
    justification = "Sequence length is greater than or equal to"
    _expected_verb = ("to have length greater than or equal to",
                      "to be greater than or equal to the length of")
    _inverse_operator = "did not have greater than or equal to the length"

    def __init__(self, sequence, length, **kwargs):
        super().__init__(SandboxedValue(sequence), SandboxedValue(length), **kwargs)

    def condition(self, sequence, length):
        """ Tests if the needle is not in the haystack """
        return len(sequence.value) < length.value


class assert_is_instance(RuntimeAssertionFeedback):
    """
    Determine if the ``obj`` is an instance of ``cls``
    """
    justification = "Object is not an instance of class"
    _expected_verb = ("to be an instance of", "to be the type of")
    _inverse_operator = "is not an instance of"

    def __init__(self, obj, cls, **kwargs):
        super().__init__(SandboxedValue(obj), SandboxedValue(cls), **kwargs)

    def condition(self, obj, cls):
        """ Tests if the left and right are equal """
        value = cls.value
        if value == int or value == float:
            value = (int, float)
        return not isinstance(obj.value, value)


class assert_not_is_instance(RuntimeAssertionFeedback):
    """
    Determine if the ``obj`` is an instance of ``cls``
    """
    justification = "Object is an instance of class"
    _expected_verb = ("to not be an instance of", "to not be the type of")
    _inverse_operator = "is an instance of"

    def __init__(self, obj, cls, **kwargs):
        super().__init__(SandboxedValue(obj), SandboxedValue(cls), **kwargs)

    def condition(self, obj, cls):
        """ Tests if the left and right are equal """
        value = cls.value
        if value == int or value == float:
            value = (int, float)
        return isinstance(obj.value, value)


def type_to_pedal_type(expected_type):
    evaluated_expected_type = evaluate(expected_type) if isinstance(expected_type, str) else expected_type
    expected_pedal_type = normalize_type(evaluated_expected_type, evaluate)
    if not isinstance(expected_pedal_type, Exception):
        expected_pedal_type = expected_pedal_type.as_type()
        expected_pedal_type_name = expected_pedal_type.singular_name
    else:
        expected_pedal_type_name = ""
    return expected_pedal_type, expected_pedal_type_name


def value_to_pedal_type(value):
    if isinstance(unwrap_value(value), Exception):
        value_pedal_type = "An error"
    else:
        value_pedal_type = get_pedal_type_from_value(unwrap_value(value), evaluate)
    return value_pedal_type


class _compare_type(RuntimeAssertionFeedback):
    """
    TODO: Failing for assert_type({"test":1}, dict)
    """

    def __init__(self, value, expected_type, **kwargs):
        fields = kwargs.setdefault('fields', {})
        value_pedal_type = value_to_pedal_type(value)
        singular_name = share_sandbox_context(value_pedal_type if
                                              isinstance(value_pedal_type, str) else
                                              value_pedal_type.singular_name, value)
        expected_pedal_type, expected_pedal_type_name = type_to_pedal_type(expected_type)
        fields['value_raw'] = value
        fields['value_type'] = value_pedal_type
        fields['value_type_name'] = singular_name
        fields['expected_type_raw'] = expected_type
        fields['expected_type'] = expected_pedal_type
        fields['expected_type_name'] = expected_pedal_type_name
        super().__init__(SandboxedValue(singular_name),
                         SandboxedValue(expected_pedal_type_name), **kwargs)

    def condition(self, value, expected_type):
        """ Tests if the left and right are equal """
        value_type = self.fields['value_type']
        expected_type = self.fields['expected_type']
        return not is_subtype(value_type, expected_type)


class assert_type(_compare_type):
    """ Same as assert_is_instance, but has a slightly different wording. """
    justification = "Value is not of type"
    _expected_verb = ("to be a value of type", "to be the type of")
    _inverse_operator = "is a value of type"


class assert_not_type(_compare_type):
    justification = "Value is of type"
    _expected_verb = ("to not be a value of type", "to not be the type of")
    _inverse_operator = "is not a value of type"

    def condition(self, value, expected_type):
        return not super().condition(value, expected_type)


class assert_regex(RuntimeAssertionFeedback):
    """
    Determine if the ``regex`` matches ``text``.
    """
    justification = "Regex does not match text"
    _expected_verb = ("to be matched by the regex", "to match the text")
    _inverse_operator = "does not match the text"

    def __init__(self, regex, text, **kwargs):
        super().__init__(SandboxedValue(regex), SandboxedValue(text), **kwargs)

    def condition(self, regex, text):
        """ Tests if the regex matches the text """
        return re.search(regex.value, str(text.value)) is None


class assert_not_regex(RuntimeAssertionFeedback):
    """
    Determine if the ``regex`` does not match ``text``.
    """
    justification = "Regex matches text"
    _expected_verb = ("to not match the regex", "to not match the text")
    _inverse_operator = "matches the text"

    def __init__(self, regex, text, **kwargs):
        super().__init__(SandboxedValue(regex), SandboxedValue(text), **kwargs)

    def condition(self, regex, text):
        """ Tests if the regex does not match the text """
        return re.search(regex.value, str(text.value)) is not None


class assert_almost_equal(assert_equal):
    """ Test if the two values are almost equal; equivalent to assert_equal. """


class assert_not_almost_equal(assert_not_equal):
    """ Test if the two values are not almost equal; equivalent to
    assert_not_equal. """


class assert_output(RuntimePrintingAssertionFeedback):
    """
    Determine if the ``execution`` outputs ``text``. Uses the `==` operator to do the final comparison.
    In this case, you can think of the output as a single string with newlines, as opposed to a list
    of strings (i.e., it is retrieved with `raw_output`).

    If the `exact_strings` parameter is set to be `False`, then output is first normalized following
    this strategy:
    * Make strings lowercase
    * Remove all punctuation characters
    * Split the string by newlines into a list
    * Split each individual line by spaces (aka into words)
    * Remove all empty lines
    * Sorts the lines by default order

    So the default check will be fairly generous about checking output; as long as all the lines are
    there (in whatever order), ignoring punctuation and case, the text will be found.
    """
    justification = "Did not print the output"
    _expected_verb = "the output to be"
    _inverse_operator = "does not have the text"

    def condition(self, execution, text, exact_strings):
        """ Tests if the regex does not match the text """
        return errors(execution) or not equality_test(self.get_output(execution), str(text.value),
                                                      _exact_strings=exact_strings, _delta=None)


class assert_prints(assert_output):
    """ Deprecated version of assert_output """


class assert_not_output(RuntimePrintingAssertionFeedback):
    """
    Determine if the ``execution`` does not output ``text``. Simply the inverse of
    :py:func:`pedal.assertions.runtime.assert_output` so check the rules there for more information.
    """
    justification = "Printed the output"
    _expected_verb = "the output to not be"
    _inverse_operator = "has the text"

    def condition(self, execution, text, exact_strings):
        """ Tests if the regex does not match the text """
        return equality_test(self.get_output(execution), str(text.value),
                             _exact_strings=exact_strings, _delta=None)


class assert_output_contains(RuntimePrintingAssertionFeedback):
    """
    Determine if the ``execution`` outputs ``text`` anywhere. Unlike :py:func:`pedal.assertions.runtime.assert_output`,
    this function uses the `in` operator. If the `exact_strings` parameter is `False`, then both strings are only
    lowercased first (but the other normalization rules from `assert_output` are not applied). Can be a more flexible
    check since it just looks for whether the run of characters is ANYWHERE in the output. Remember that newlines are
    part of the output, though, so the check will not work across lines unless the `text` includes those newlines.
    """
    justification = "Did not contain the printed output"
    _expected_verb = "the output to contain"
    _inverse_operator = "does not contain the text"

    def condition(self, execution, text, exact_strings):
        """ Tests if the regex does not match the text """
        if not exact_strings:
            return str(text.value).lower() not in self.get_output(execution).lower()
        return str(text.value) not in self.get_output(execution)


class assert_not_output_contains(RuntimePrintingAssertionFeedback):
    """
    Determine if the ``execution`` outputs ``text``
    """
    justification = "Contained the printed output"
    _expected_verb = "the output to not contain"
    _inverse_operator = "contained the text"

    def condition(self, execution, text, exact_strings):
        """ Tests if the regex does not match the text """
        if not exact_strings:
            return str(text.value).lower() in self.get_output(execution).lower()
        return str(text.value) in self.get_output(execution)


class assert_output_regex(RuntimePrintingAssertionFeedback):
    """
    Determine if the ``execution`` output matches the given regex, similar to assert_output and assert_regex.
    """
    justification = "Did not print the output matching the regex"
    _expected_verb = "the output to match the regex"
    _inverse_operator = "does not match the text"

    def __init__(self, pattern, execution, exact_strings=False, **kwargs):
        super().__init__(execution, pattern,
                         exact_strings=exact_strings, **kwargs)

    def condition(self, execution, text, exact_strings):
        """ Tests if the regex does not match the text """
        return errors(execution) or re.search(str(text.value), self.get_output(execution)) is None


class assert_not_output_regex(RuntimePrintingAssertionFeedback):
    """
    Determine if the ``execution`` does not output ``text``. Simply the inverse of
    :py:func:`pedal.assertions.runtime.assert_output` so check the rules there for more information.
    """
    justification = "Printed the output matching the regex"
    _expected_verb = "the output to not match the regex"
    _inverse_operator = "matches the text"

    def __init__(self, pattern, execution, exact_strings=False, **kwargs):
        super().__init__(execution, pattern,
                         exact_strings=exact_strings, **kwargs)

    def condition(self, execution, text, exact_strings):
        """ Tests if the regex does not match the text """
        return re.search(str(text.value), self.get_output(execution)) is not None


class assert_has_attr(RuntimeAssertionFeedback):
    """
    Determine if the ``object`` has the ``name``
    """
    justification = "Contained the attribute"
    _expected_verb = "the object to contain"
    _inverse_operator = "did not contain"

    def condition(self, obj, attr, exact_strings):
        """ Tests if the regex does not match the text """
        return hasattr(obj.value, attr.value)


class assert_has_variable(RuntimeAssertionFeedback):
    """
    Determine if the student's data has the name.
    """
    _expected_verb = "to contain the variable"
    _inverse_operator = "does not contain the variable"

    def __init__(self, sandbox, variable_name, **kwargs):
        super().__init__(SandboxedValue(sandbox),
                         ExactValue(variable_name), **kwargs)

    def format_assertion(self, sandbox, variable_name, contexts):
        variable_name = variable_name.value
        return f"The variable {variable_name} was not created."


    def condition(self, sandbox, variable_name):
        sandbox = sandbox.value
        variable_name = variable_name.value
        if isinstance(sandbox, Sandbox):
            return variable_name not in sandbox.data
        elif isinstance(sandbox, dict):
            return variable_name not in sandbox
        else:
            return True


class assert_has_function(RuntimeAssertionFeedback):
    """
    Determine if the student's code has the function.
    """
    _expected_verb = "to contain the function"
    _inverse_operator = "does not contain the function"

    def __init__(self, sandbox, function_name, **kwargs):
        super().__init__(SandboxedValue(sandbox),
                         ExactValue(function_name), **kwargs)

    def format_assertion(self, sandbox, function_name, contexts):
        sandbox = sandbox.value
        if isinstance(sandbox, Sandbox):
            return "Sandbox does not contain the function."
        else:
            return "The result does not contain the function."

    def condition(self, sandbox, function_name):
        sandbox = sandbox.value
        function_name = function_name.value
        if isinstance(sandbox, Sandbox):
            function = sandbox.data.get(function_name, None)
        elif isinstance(sandbox, dict):
            function = sandbox.get(function_name, None)
        elif hasattr(sandbox, function_name):
            function = getattr(sandbox, function_name)
        else:
            function = None
        return not callable(function)

# TODO: This one is at Runtime, but is not an assertion... Should these be "tests"?


class ensure_coverage(AssertionFeedback):
    """
    Verifies that the most recent executed and traced student code has
    ``at_least`` the given ratio of covered (executed) lines.

    Args:
        at_least (float): The ratio of covered lines. A value of 1.0 is all
            lines covered, 0.0 is no lines covered, and .5 is half the lines
            covered.
    """
    title = "You Must Test Your Code"
    message_template = ("Your code coverage is not adequate. You must cover at "
                        "least {at_least_message}% of your code to receive "
                        "feedback. So far, you have only covered "
                        "{coverage_message}%.")

    def __init__(self, at_least=.5, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.setdefault('fields', {})
        fields['at_least'] = at_least
        fields['at_least_message'] = str(int(round(100*at_least)))
        unexecuted_lines, coverage = check_coverage(report)
        fields['unexecuted_lines'] = unexecuted_lines
        fields['coverage'] = coverage
        fields['coverage_message'] = str(int(round(100*coverage)))
        super().__init__(coverage, at_least, **kwargs)

    def condition(self, coverage, at_least):
        return coverage < at_least


class ensure_called_uniquely(AssertionFeedback):
    """
    Verifies that the most recent executed and traced student code has
    ``at_least`` called the given function uniquely that number of times.
    In other words, it prevents students from calling the same function repeatedly
    WITHOUT changing the arguments.

    TODO: Allow instructor to ignore certain collection of arguments
    TODO: Report how many calls it has seen so far.

    Args:
        function_name (str): The name of the function to check.
        at_least (int): The number of calls that have to have unique arguments
            between them.
        ignore (set[tuple]): A sequence of argument sets to ignore.
        why_ignored (str): If you want to explain why you are ignoring some of
            the tests, you can provide some text here. For example,
            `" because they overlap with examples you were given"`.
    """
    title = "You Must Test Your Code"
    message_template = ("You have not tested the function {function_name} enough. "
                        "You should test it at least {at_least} times. Each time you"
                        " test it, you should be using a new set of arguments."
                        " So far, you have called it {total_calls} times in total and"
                        " {unique_calls} times distinctively{instructor_ignore_message}.")

    def __init__(self, function_name, at_least=1, ignore=None, why_ignored="", **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.setdefault('fields', {})
        fields['function_name'] = function_name
        fields['at_least'] = at_least
        if ignore is None:
            ignore = set()
        else:
            ignore = set(ignore)
        fields['ignore'] = ignore
        calls = get_call_arguments(function_name, report)
        unique_calls = set([tuple(map(repr, args.values())) for args in calls])
        fields['instructor_ignored'] = instructor_ignored = len(unique_calls & ignore)
        if instructor_ignored:
            fields['instructor_ignore_message'] = f" (but your instructor did not count {instructor_ignored} of the tests{why_ignored})"
        else:
            fields['instructor_ignore_message'] = ""
        fields['total_calls'] = len(calls)
        unique_call_count = len(unique_calls - ignore)
        fields['unique_calls'] = unique_call_count
        super().__init__(unique_call_count, at_least, **kwargs)

    def condition(self, unique_call_count, at_least):
        return unique_call_count < at_least


@CompositeFeedbackFunction(function_not_available, name_is_not_a_function)
def ensure_function_callable(name, **kwargs):
    report = kwargs.get("report", MAIN_REPORT)
    values = get_student_data(report)
    # 2.1. Does the name exist in the values?
    if name not in values:
        return function_not_available(name, **kwargs)
    function_value = values[name]
    # 2.2. Is the name bound to a callable?
    if not callable(function_value):
        return name_is_not_a_function(name)

# Alias conventional camel-case names to our functions

assertEqual = assert_equal
assertNotEqual = assert_not_equal
assertLess = assert_less
assertLessEqual = assert_less_equal
assertGreater = assert_greater
assertGreaterEqual = assert_greater_equal
assertLengthEqual = assert_length_equal
assertLengthNotEqual = assert_length_not_equal
assertLengthLess = assert_length_less
assertLengthLessEqual = assert_length_less_equal
assertLengthGreater = assert_length_greater
assertLengthGreaterEqual = assert_length_greater_equal
assertIn = assert_in
assertNotIn = assert_not_in
assertIs = assert_is
assertIsNot = assert_is_not
assertIsNone = assert_is_none
assertIsNotNone = assert_is_not_none
assertTrue = assert_true
assertFalse = assert_false
assertIsInstance = assert_is_instance
assertIsNotInstance = assert_not_is_instance
assertAlmostEqual = assert_equal
assertNotAlmostEqual = assert_not_equal
assertRegex = assert_regex
assertNotRegex = assert_not_regex
assertPrints = assert_prints
assertOutput = assert_output
assertNotOutput = assert_not_output
assertOutputContains = assert_output_contains
assertNotOutputContains = assert_not_output_contains
assertHasAttr = assert_has_attr
assertHasFunction = assert_has_function
assertHasVariable = assert_has_variable
assertType = assert_type
assertNotType = assert_not_type



import unittest
import os
import sys
import json
import re

from pedal.core.feedback import Feedback
from pedal.core.final_feedback import FinalFeedback
from pedal.utilities.system import IS_AT_LEAST_PYTHON_311

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from pedal.core.commands import (clear_report, set_success, gently, explain, give_partial, suppress,
                                 contextualize_report, get_all_feedback, feedback)
from pedal.source import set_source, next_section, verify_section, verify, separate_into_sections
from pedal.tifa import tifa_analysis
from pedal.resolvers import simple, sectional, statistics
import pedal.sandbox.commands as commands
from tests.execution_helper import Execution, SUCCESS_TEXT, ExecutionTestCase, SUCCESS_MESSAGE


class TestResolver(ExecutionTestCase):

    def test_do_nothing(self):
        clear_report()
        final = simple.resolve()
        self.assertTrue(final.success)
        self.assertEqual(final.message, SUCCESS_TEXT)

    def test_gently(self):
        clear_report()
        gently('You should always create unit tests.')
        final = simple.resolve()
        self.assertFalse(final.success)
        self.assertEqual(final.message, 'You should always create unit tests.')

    def test_gently_order(self):
        clear_report()
        gently('A great and exciting message!')
        gently('A boring message that we should not show.')
        final = simple.resolve()
        self.assertFalse(final.success)
        self.assertEqual(final.message, 'A great and exciting message!')

    def test_set_success(self):
        clear_report()
        set_success()
        final = simple.resolve()
        self.assertTrue(final.success)
        self.assertEqual(final.message, 'Great work!')

    def test_gently_and_set_success(self):
        clear_report()
        gently("What have you done?")
        set_success()
        final = simple.resolve()
        self.assertFalse(final.success)
        self.assertEqual(final.message, 'What have you done?')

    def test_explain(self):
        # Tifa < Explain
        with Execution('1+""') as e:
            explain("You cannot add those.")
        self.assertEqual(e.final.message, "You cannot add those.")
        # Tifa > Gently
        with Execution('1+""') as e:
            gently("You cannot add those.")
        self.assertEqual(e.final.title, "Incompatible types")

    def test_hidden_error(self):
        clear_report()
        contextualize_report('import pedal')
        verify()
        tifa_analysis()
        final = simple.resolve()
        self.assertNotEqual("No errors reported.", final.message)

    def test_unmessaged_tifa(self):
        clear_report()
        contextualize_report('import random\nrandom')
        verify()
        tifa_analysis()
        final = simple.resolve()
        self.assertEqual(SUCCESS_MESSAGE, final.title+"\n"+final.message)

    def test_partials(self):
        with Execution('0') as e:
            gently("You were incorrect.")
            give_partial(.1, message="You had a zero in your code.")
            give_partial(.1, message="You looped correctly.")
        self.assertFeedback(e, "Instructor Feedback\nYou were incorrect.")
        self.assertEqual(.2, e.final.score)
        self.assertFalse(e.final.success)

        with Execution('0') as e:
            give_partial(.1, message="You had a zero in your code.")
            give_partial(.1, message="You looped correctly.")
            gently("Okay but you still only wrote 0.")
        self.assertEqual(e.final.message, "Okay but you still only wrote 0.")
        self.assertEqual(e.final.score, .2)
        self.assertFalse(e.final.success)

        with Execution('0') as e:
            give_partial(.1, message="You had a zero in your code.")
            give_partial(.1, message="You looped correctly.")
            set_success()
        self.assertEqual(e.final.message, "Great work!")
        self.assertEqual(e.final.score, 1.2)
        self.assertTrue(e.final.success)

    def test_analyzer_suppression(self):
        clear_report()
        contextualize_report('1+"Hello"')
        verify()
        tifa_analysis()
        commands.run()
        suppress("analyzer")
        final = simple.resolve()
        self.assertEqual("runtime", final.category)
        self.assertEqual("Type Error", final.title)

    def test_runtime_suppression(self):
        clear_report()
        contextualize_report('import json\njson.loads("0")+"1"')
        verify()
        tifa_analysis()
        commands.run()
        suppress("Runtime")
        final = simple.resolve()
        self.assertEqual(Feedback.CATEGORIES.COMPLETE, final.category)
        self.assertEqual(SUCCESS_TEXT, final.message)

    def test_success(self):
        clear_report()
        contextualize_report('a=0\na')
        verify()
        tifa_analysis()
        set_success()
        final = simple.resolve()
        self.assertEqual(Feedback.CATEGORIES.COMPLETE, final.category)
        self.assertEqual("Complete", final.title)
        self.assertEqual("Great work!", final.message)

    def test_success_suppression(self):
        clear_report()
        contextualize_report('a=0\na')
        verify()
        tifa_analysis()
        set_success()
        suppress(label='set_success')
        final = simple.resolve()
        self.assertEqual(Feedback.CATEGORIES.COMPLETE, final.category)
        self.assertEqual(SUCCESS_MESSAGE, final.title+"\n"+final.message)

    def test_empty(self):
        clear_report()
        contextualize_report('    ')
        verify()
        tifa_analysis()
        commands.run()
        final = simple.resolve()
        self.assertEqual(Feedback.CATEGORIES.SYNTAX, final.category)
        self.assertEqual("No Source Code", final.title)
        self.assertEqual("Source code file is blank.", final.message)

    def test_gently_vs_runtime(self):
        # Runtime > Gently
        clear_report()
        contextualize_report('import json\njson.loads("0")+"1"')
        verify()
        tifa_analysis()
        commands.run()
        gently("I have a gentle opinion, but you don't want to hear it.")
        final = simple.resolve()
        print(final.label)
        self.assertEqual(Feedback.CATEGORIES.RUNTIME, final.category)

        # Runtime < Explain
        clear_report()
        contextualize_report('import json\njson.loads("0")+"1"')
        verify()
        tifa_analysis()
        commands.run()
        explain("LISTEN TO ME")
        final = simple.resolve()
        self.assertEqual(Feedback.CATEGORIES.INSTRUCTOR, final.category)

    def test_input(self):
        with Execution('input("Type something:")') as e:
            pass
        self.assertNotEqual(Feedback.CATEGORIES.RUNTIME, e.final.category)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_input_with_conversion(self):
        with Execution('float(input("Type something:"))') as e:
            pass
        self.assertNotEqual(Feedback.CATEGORIES.RUNTIME, e.final.category)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_sectional_error(self):
        clear_report()
        contextualize_report('a=0\n##### Part 1\nprint("A")\n##### Part 2\nsyntax error')
        separate_into_sections(independent=True)
        # Part 0
        verify()
        commands.clear_sandbox()
        commands.run()
        # Part 1
        next_section()
        verify()
        commands.clear_sandbox()
        commands.run()
        give_partial(.2)
        # Part 2
        next_section()
        verify()
        commands.clear_sandbox()
        commands.run()
        give_partial(.2)
        # Resolve everything
        finals = sectional.resolve()
        self.assertEqual("""# Global
FeedbackSourceSection
Feedback separated into groups
# 1
Complete
Great work!
# 2
Syntax Error
Bad syntax on line 5.

The traceback was:
Line 5 of file answer.py
    syntax error
""" + ('           ^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") + """
Invalid syntax.
Suggestion: Check line 5, the line before it, and the line after it. Ignore blank lines.""",
                         "\n".join(f"# {g.section_number if g is not None else 'Global'}\n{f.title}\n{f.message}"
                                   for g, f in finals.items()))

    def test_sectional_success(self):
        clear_report()
        contextualize_report('a=0\n##### Part 1\nprint("A")\n##### Part 2\nprint("B")')
        separate_into_sections(independent=True)
        # Part 0
        verify()
        commands.clear_sandbox()
        commands.run()
        # Part 1
        next_section()
        verify()
        commands.clear_sandbox()
        commands.run()
        give_partial(.2)
        # Part 2
        next_section()
        verify()
        commands.clear_sandbox()
        commands.run()
        give_partial(.2)
        # Resolve everything
        finals = sectional.resolve()
        self.assertEqual("""# Global
FeedbackSourceSection
Feedback separated into groups
# 1
Complete
Great work!
# 2
Complete
Great work!""",
                         "\n".join(f"# {g.section_number if g is not None else 'Global'}\n{f.title}\n{f.message}"
                                   for g, f in finals.items()))

    def test_attribute_error(self):
        with Execution('"".unsafe()', run_tifa=False) as e:
            pass
        self.assertEqual("Attribute Error", e.final.title)

    def test_combining_scores_complex(self):
        clear_report()
        contextualize_report('a=0\nprint(a)')
        # These are added
        feedback(activate=True, valence=1, score="+4%", category='instructor')
        # These are skipped
        feedback(activate=False, valence=1, score="+5%", category='instructor')
        # These are skipped
        feedback(activate=True, valence=-1, score="+8%", category='instructor')
        # These are added
        feedback(activate=False, valence=-1, score="+7%", category='instructor')
        # Calculate final result
        final = simple.resolve()
        self.assertEqual(.11, final.score)

    def test_suppress_specific_function(self):
        with Execution("""
def x():
    pass
def y():
    pass
def z():
    pass
z""", run_tifa=True) as e:
            suppress("analyzer", "unused_variable", {"name": "x"})
        self.assertFeedback(e, """Unused Variable
The function y was given a definition on line 4, but was never used after that.""")

    def test_suppress_specific_function_twice(self):
        with Execution("""
def x():
    pass
def y():
    pass
def z():
    pass
def q():
    pass
z""", run_tifa=True) as e:
            suppress("analyzer", "unused_variable", {"name": "x"})
            suppress("analyzer", "unused_variable", {"name": "y"})
        self.assertFeedback(e, """Unused Variable
The function q was given a definition on line 8, but was never used after that.""")

    def test_suppress_all_individually(self):
        with Execution("""
p = 0
def x():
    pass
def y():
    pass
""", run_tifa=True) as e:
            suppress("analyzer", "unused_variable", {"name": "x"})
            suppress("analyzer", "unused_variable", {"name": "y"})
        self.assertFeedback(e, """Unused Variable
The variable p was given a value on line 2, but was never used after that.""")

        with Execution("""
def x():
    pass
def y():
    pass
p = 0
    """, run_tifa=True) as e:
            suppress("analyzer", "unused_variable", {"name": "x"})
            suppress("analyzer", "unused_variable", {"name": "y"})
        self.assertFeedback(e, """Unused Variable
The variable p was given a value on line 6, but was never used after that.""")

    def test_stats_resolver(self):
        clear_report()
        contextualize_report('a=0\na+""')
        # These are added
        verify()
        tifa_analysis()
        commands.run()
        # Calculate final result
        final = statistics.resolve(clean=True)
        # Get expected
        with open('output_files/stats_feedback_test_1.json') as f:
            expected = f.read()
        final = re.sub(r"\s*\^\^+\\n", '', final)
        self.assertEqual(final, expected)


if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Tests related to checking function definitions
"""

import unittest
from dataclasses import dataclass
from textwrap import dedent

from pedal import unit_test, block_function
from pedal.assertions.static import *
from pedal.assertions.commands import *
from pedal.assertions.runtime import *
from pedal.types.new_types import DictType, LiteralStr, StrType
from pedal.utilities.system import IS_AT_LEAST_PYTHON_39
from tests.execution_helper import Execution, ExecutionTestCase, SUCCESS_MESSAGE


class TestAssertionsDataclasses(ExecutionTestCase):
    def test_dataclasses_passes_runtime(self):
        with Execution("""
from dataclasses import dataclass

@dataclass
class Dog:
    name: str
    age: int

x = Dog("Ada", 4)
print(x)""") as e:
            @dataclass
            class Dog:
                name: str
                age: int
            self.assertFalse(check_dataclass_instance(evaluate('x'), Dog))
            #self.assertFalse(assert_dataclass_fields())
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_dataclasses_fails_runtime(self):
        with Execution("""
from dataclasses import dataclass

@dataclass
class Dog:
    name: str
    age: int

x = Dog("Ada", 4)
print(x)
x.age = "Apple"
print(x)""") as e:
            @dataclass
            class Dog:
                name: str
                age: int
            self.assertTrue(check_dataclass_instance(evaluate('x'), Dog))
            #self.assertFalse(assert_dataclass_fields())
        self.assertFeedback(e, """Wrong Fields Type
The dataclass named Dog has a field named age that is a string, but should be an integer.""")

    def test_dataclasses_fails_runtime_close(self):
        with Execution("""
from dataclasses import dataclass

@dataclass
class Dog:
    name: str
    ag: int

x = Dog("Ada", 4)
print(x)""") as e:
            @dataclass
            class Dog:
                name: str
                age: int

            self.assertTrue(check_dataclass_instance(evaluate('x'), Dog))
            # self.assertFalse(assert_dataclass_fields())
        self.assertFeedback(e, """Unknown Field
The dataclass named Dog had a field named ag but that field is not supposed to be there. Are you sure you got the name right?""")

    def test_dataclasses_fails_runtime_error(self):
        with Execution("""
from dataclasses import dataclass

@dataclass
class Dog:
    name: str
    ag: int

x = Dog("Ada")
print(x)""") as e:
            @dataclass
            class Dog:
                name: str
                age: int

            self.assertTrue(check_dataclass_instance(evaluate('x'), Dog))
            # self.assertFalse(assert_dataclass_fields())
        self.assertFeedback(e, """Incorrect Arity
The constructor function Dog was given the wrong number of arguments. You should have had 2 arguments, but instead you had 1 arguments.""")

    @unittest.skipIf(not IS_AT_LEAST_PYTHON_39, "Cannot subscript literals in Python 3.7")
    def test_weird_forecast_issue(self):
        @dataclass
        class WeatherOptions:
            raining: bool
            cloudy: bool
            snowing: bool

        @dataclass
        class Measurement:
            amount: int
            automatic: bool

        @dataclass
        class Report:
            temperature: int
            rainfall: list[Measurement]
            weather: WeatherOptions

        @dataclass
        class Forecast:
            when: str
            where: str
            reports: list[Report]

        CLASSES = [Forecast, Report, Measurement, WeatherOptions]
        NAMES = ["Forecast", "Report", "Measurement", "WeatherOptions"]
        main_program = dedent("""from dataclasses import dataclass
from bakery import assert_equal

@dataclass
class Measurement:
    amount: int
    automatic: bool

@dataclass
class WeatherOptions:
    raining:bool
    cloudy: bool
    snowing: bool

@dataclass
class Report:
    temperature: int
    rainfall: list[Measurement]
    weather: WeatherOptions

@dataclass
class Forecast:
    when: str
    where: str
    reports: list[Report]

def total_rainfall(forecasts: list[Forecast]) -> int:
    total = 0
    for forecast in forecasts:
        for report in forecast.reports:
            for rain in report.rainfall:
                total = total + rain.amount
    return total



List1 = [
    Forecast("Tuesday", "Dover", [
        Report(20,
               [Measurement(30, True)], 
               WeatherOptions(True, True, False))
    ])
]
List2 = [
    Forecast("Wednesday", "Newark", [
        Report(30,
               [Measurement(40, True), Measurement(25, True)],
               WeatherOptions(True, True, False))
    ])
]
assert_equal(total_rainfall(List1), 30)
assert_equal(total_rainfall(List2), 65)""")
        with Execution(main_program):
            for name, dc in zip(NAMES, CLASSES):
                self.assertFalse(ensure_dataclass(dc, priority='instructor'))
                self.assertFalse(assert_is_instance(evaluate(name), type))
                #
            # self.assertFalse(assert_dataclass_fields())
        #self.assertFeedback(e, """Incorrect Arity
        #The constructor function Dog was given the wrong number of arguments. You should have had 2 arguments, but instead you had 1 arguments.""")

if __name__ == '__main__':
    unittest.main(buffer=False)
